/********************************************************************

    backend.cpp

********************************************************************/


#include "backend.h"


#include "../machine/vmcodes.h"



//--------------------------------------
  void BackEnd::init ( int _numThreads )
//--------------------------------------
{
  links.clear();
  calls.clear();

  numThreads  = _numThreads;
  varBase     = numThreads * SIZEOF_THREAD;
  varOffset   = 0;
  constOffset = 1 + (numThreads * SIZEOF_THREAD_INIT) / 2;
  codeOffset  = 0;
  instructionNo = -1;
}


//-------------------------
  void BackEnd::alignVar ()
//-------------------------
{
  varOffset++;
  varOffset &= ~1;
}


//-----------------------------------------------------
  void BackEnd::allocateVariable ( Variable* variable )
//-----------------------------------------------------
{
  Type* type = dynamic_cast<Type*>(variable->getType());

  if ( type->getSize() > 1 )
    alignVar();

  variable->setOffset(varBase + varOffset);
  varOffset += variable->getSize();
}


//------------------------------------------------------
  unsigned short BackEnd::storeConst ( unsigned char b )
//------------------------------------------------------
{
  unsigned short offset = constOffset;
  constants[constOffset] = b;
  constOffset++;
  return offset;
}


//----------------------------------------------
  unsigned short BackEnd::storeConst ( short i )
//----------------------------------------------
{
  unsigned short offset = constOffset;
  *(short*)(constants + constOffset) = i;
  constOffset++;
  return offset;
}


//---------------------------------------------
  unsigned short BackEnd::storeConst ( long l )
//---------------------------------------------
{
  unsigned short offset = constOffset;
  *(long*)(constants + constOffset) = l;
  constOffset += 2;
  return offset;
}


//-----------------------------------------------
  unsigned short BackEnd::storeConst ( double f )
//-----------------------------------------------
{
  unsigned short offset = constOffset;
  *(double*)(constants + constOffset) = f;
  constOffset += 4;
  return offset;
}


//---------------------------------------------------------
  unsigned short BackEnd::storeConst ( Constant* constant )
//---------------------------------------------------------
{
  NumericType* type = dynamic_cast<NumericType*>(constant->getType());
  unsigned short offset;

  switch ( type->getForm() )
  {
    case NumericType::BYTE  : offset = storeConst((unsigned char) constant->getValue()); break;
    case NumericType::INT   : offset = storeConst((short) constant->getValue()); break;
    case NumericType::LONG  : offset = storeConst((long) constant->getValue()); break;
    case NumericType::FLOAT : offset = storeConst((double) constant->getValue()); break;
  }

  constant->setOffset(offset);
  return offset;
}


//----------------------------------------------------------------
  unsigned short BackEnd::storeConst ( ArrayConstant* arrayconst )
//----------------------------------------------------------------
{
  NumericType* type = dynamic_cast<NumericType*>(arrayconst->getType());
  NumericType::FORM form = type->getForm();

  unsigned short offset = constOffset;
  arrayconst->setOffset(offset);
 
  for ( int i = 0; i < arrayconst->getNumArrayElements(); i++)
  {
    switch ( form )
    {
      case NumericType::BYTE  : storeConst((unsigned char) arrayconst->getValue(i)); break;
      case NumericType::INT   : storeConst((short) arrayconst->getValue(i)); break;
      case NumericType::LONG  : storeConst((long) arrayconst->getValue(i)); break;
      case NumericType::FLOAT : storeConst((double) arrayconst->getValue(i)); break;
    }
  }
  return offset;
}


//---------------------------------------------------------------
  unsigned short BackEnd::storeConst ( StringConstant* strconst )
//---------------------------------------------------------------
{
  unsigned short offset = constOffset;
  strconst->setOffset(offset);

  const int MAX_LENGTH = 30;  
  int length = strconst->getString().length() < MAX_LENGTH ? strconst->getString().length() : MAX_LENGTH;
  
  unsigned char* ptr = (unsigned char*) (constants + offset);
  ptr[0] = (unsigned char)length;

  int i;

  for ( i = 1; i <= length; i++ )
    ptr[i] = (unsigned char) strconst->getString()[i-1]; 

  constOffset += (i + 1) / 2;

  return offset;
}


//-------------------------------------------------------------------------
  unsigned short BackEnd::storeConst ( StringArrayConstant* strarrayconst )
//-------------------------------------------------------------------------
{
  const int MAX_LENGTH = 30;  

  unsigned short offset = constOffset;
  short adr = offset + strarrayconst->getNumArrayElements();
  int i, k;
  int length = 0;
  unsigned char* ptr;

  for ( i = 0; i < strarrayconst->getNumArrayElements(); i++ )
  {
    const std::string& str = strarrayconst->getString(i);
    storeConst(adr);
    length = str.length();
    length = length > MAX_LENGTH ? MAX_LENGTH : length;


    ptr = (unsigned char*) (constants + adr);
    ptr[0] = (unsigned char)length;

    for ( k = 1; k <= length; k++ )
      ptr[k] = (unsigned char) str[k-1]; 

    adr += (length + 2) / 2;
  }

  constOffset = adr;
  return offset;
}


//-------------------------------------------------------------------------------------
  int BackEnd::compile ( IntermediateCodeSequence* imcodeseq, const std::string& name )
//-------------------------------------------------------------------------------------
{
  int offset  = codeOffset;
  links[name] = (unsigned short)offset;

  marks.clear();
  jumps.clear();

  while ( !types.empty() ) 
    types.pop();

  imcodeseq->finish();

  if ( imcodeseq->getLocalMemSize() )
  {
    unsigned short localMemSize = (unsigned short)imcodeseq->getLocalMemSize();
    localMemSize++;
    localMemSize &= ~1;
    instructionNo = -1;
    putCode(VM_INC_SP);
    putCode(localMemSize);
  }

  for ( imcodeseq->rewind(); !imcodeseq->isLastCode(); imcodeseq->gotoNextCode() )
  {
    IntermediateCode& imCode = imcodeseq->getCode();
    
    switch ( imCode.opcode ) 
    {
      case IntermediateCode::NOP:
        break;
      
      case IntermediateCode::LOAD     : put_LOAD_code(imCode); break;
      case IntermediateCode::R_INC_LOAD :
      case IntermediateCode::R_LOAD   : put_R_LOAD_code(imCode); break;
      case IntermediateCode::I_LOAD   : put_ILOAD_code(imCode.getValue()); break;
      case IntermediateCode::I_LOAD_F : put_ILOAD_F_code(imcodeseq->getCodeEnclosedFloat()); break;
      case IntermediateCode::INC_LOAD : put_INC_LOAD_code(imCode); break;
	    case IntermediateCode::K_STORE  :
      case IntermediateCode::STORE    : put_STORE_code(imCode); break;
      case IntermediateCode::R_STORE  : put_R_STORE_code(imCode); break;
      case IntermediateCode::CALL     : put_CALL_code(imCode); break;
      case IntermediateCode::RETURN   : put_RETURN_code(imCode); break;
      case IntermediateCode::WIPESTACK : put_WIPESTACK_code(imCode); break;
      case IntermediateCode::LOAD_PTR : put_LOAD_PTR_code(imCode); break;

      case IntermediateCode::INLINE : putCode((unsigned short)imCode.getID()); break;

      case IntermediateCode::HALT   : putCode(VM_LOAD_IMMEDIATE_INT); putCode(0);
      case IntermediateCode::RUN    : putCode((unsigned short)(VM_RUN | (getThreadID(imCode) << 8))); break;
      case IntermediateCode::RESUME : putCode((unsigned short)(VM_RESUME | (getThreadID(imCode) << 8))); break;
      case IntermediateCode::RESET  : putCode((unsigned short)(VM_RESET | (getThreadID(imCode) << 8))); break;

      case IntermediateCode::SLEEP  : putCode(VM_SLEEP); break;

      case IntermediateCode::CAPTURE : 

        putCode(VM_CAPTURE);
        break;

      case IntermediateCode::RELEASE : putCode(VM_RELEASE); break;
      case IntermediateCode::YIELD   : putCode(VM_YIELD); break;
      case IntermediateCode::QUIT    : putCode(VM_QUIT); break;

      case IntermediateCode::CAST :  put_CAST_code(imCode); break;

      case IntermediateCode::MARK :  marks[imCode.getID()] = (unsigned short)codeOffset; break;

      case IntermediateCode::GOTO :   
      
        putCode(VM_BRANCH); 
        putBranchOffset(imCode); 
        break; 

      case IntermediateCode::BRANCH_IF_ZERO :

        putCode(VM_BRANCH_IF_ZERO); 
        putBranchOffset(imCode); 
        break; 

      case IntermediateCode::NOT : put_NOT_code(); break;
      case IntermediateCode::NEG : put_NEG_code(); break;

      case IntermediateCode::ADD : putCalcCode_A(VM_ADD); break;
      case IntermediateCode::SUB : putCalcCode_A(VM_SUB); break; 
      case IntermediateCode::MUL : putCalcCode_A(VM_MUL); break; 
      case IntermediateCode::DIV : putCalcCode_A(VM_DIV); break;
      case IntermediateCode::MOD : putCalcCode_A(VM_MOD); break;
      case IntermediateCode::SHL : putCalcCode_A(VM_SHL); break;
      case IntermediateCode::SHR : putCalcCode_A(VM_SHR); break;
      case IntermediateCode::ASR : putCalcCode_A(VM_ASR); break;
      case IntermediateCode::ROL : putCalcCode_A(VM_ROL); break;
      case IntermediateCode::ROR : putCalcCode_A(VM_ROR); break;

      case IntermediateCode::EQUAL      : putCalcCode_B(VM_EQU); break;
      case IntermediateCode::NOTEQUAL   : putCalcCode_B(VM_NEQU); break;
      case IntermediateCode::HIGHER     : putCalcCode_B(VM_HI); break;
      case IntermediateCode::HIGHERSAME : putCalcCode_B(VM_HIS); break;
      case IntermediateCode::LOWER      : putCalcCode_B(VM_LO); break;
      case IntermediateCode::LOWERSAME  : putCalcCode_B(VM_LOS); break;

      case IntermediateCode::AND  : putCalcCode_C(VM_AND); break;
      case IntermediateCode::NAND : putCalcCode_C(VM_NAND); break;
      case IntermediateCode::OR   : putCalcCode_C(VM_OR); break;
      case IntermediateCode::NOR  : putCalcCode_C(VM_NOR); break;
      case IntermediateCode::XOR  : putCalcCode_C(VM_XOR); break;
      case IntermediateCode::NXOR : putCalcCode_C(VM_NXOR); break;

      case IntermediateCode::STR_CLEAR : putCode(VM_STRING | (STRING_CLEAR << 8)); break;
      case IntermediateCode::STR_PUTCH : putCode(VM_STRING | (STRING_PUT_CHAR << 8)); break;

      case IntermediateCode::STR_PUTSTR : putCode(VM_STRING | (STRING_PUT_STRING << 8)); break;

      case IntermediateCode::STR_PUTISTR : put_STR_PUTISTR_code(imcodeseq->getCodeEnclosedString()); break;
      case IntermediateCode::STR_PUTCSTR : put_STR_PUTCSTR_code(imCode); break;

      case IntermediateCode::STR_PUTVAL : break;

      case IntermediateCode::ENTER_INSTRUCTION : instructionNo = (long) imCode.getID(); break;

      default : break;
    }
  }

  fillInBranchOffsets ();

  return offset;
}


//-----------------------
  void BackEnd::finish ()
//-----------------------
{
  fillInCallOffsets();
  *(unsigned short*)constants = (unsigned short)numThreads;
}


//-------------------------------------------------------------------------------------
  void BackEnd::setThreadInitData ( int iThread, unsigned short codeOffset, 
                                    unsigned short memoryRequirements,
                                    unsigned short localVarRequirements,
                                    unsigned char priority )
//-------------------------------------------------------------------------------------
{
  unsigned short* ptr = constants + 1 + (iThread * SIZEOF_THREAD_INIT) / 2;

  alignVar();
  unsigned short dataOffset = varBase + varOffset;

  memoryRequirements++;
  memoryRequirements &= ~1;
  localVarRequirements++;
  localVarRequirements &= ~1;

  varOffset += memoryRequirements;

  *ptr++ = codeOffset;
  *ptr++ = dataOffset;
  *ptr++ = dataOffset + localVarRequirements;
  *ptr   = priority;
}


//--- protected support routines  -----------------------------------


//----------------------------------------------------------------
  unsigned short BackEnd::getThreadID ( IntermediateCode& imcode )
//----------------------------------------------------------------
{
  if ( imcode.getObject() )
  {
    Thread* thread = dynamic_cast<Thread*>(imcode.getObject());
    return (unsigned short) thread->getID();
  }
  else
    return 0;
}


//-----------------------------
  int BackEnd::getCalcTypeID ()
//-----------------------------
{
  NumericType::FORM atype, btype;

  btype = types.top(); types.pop();
  atype = types.top(); types.pop();

  switch ( atype )
  {
    case NumericType::BYTE :
    case NumericType::INT :
      
      switch ( btype )
      {
        case NumericType::BYTE :
        case NumericType::INT   : return CALC_INT_INT;
        case NumericType::LONG  : return CALC_INT_LONG;
        default                 : return CALC_INT_FLOAT;
      }

    case NumericType::LONG :

      switch ( btype )
      {
        case NumericType::BYTE :
        case NumericType::INT   : return CALC_LONG_INT;
        case NumericType::LONG  : return CALC_LONG_LONG;
        default                 : return CALC_LONG_FLOAT;
      }

    default :

      switch ( btype )
      {
        case NumericType::BYTE :
        case NumericType::INT   : return CALC_FLOAT_INT;
        case NumericType::LONG  : return CALC_FLOAT_LONG;
        default                 : return CALC_FLOAT_FLOAT;
      }
  }
}


//----------------------------------
  void BackEnd::fillInCallOffsets ()
//----------------------------------
{
  std::map<unsigned short, std::string>::iterator iter;

  for ( iter = calls.begin(); iter != calls.end(); ++iter )
  {
    unsigned short fillInPos = (*iter).first;
    std::string& calledFunctionName = (*iter).second;
    unsigned short offset = links[calledFunctionName];
    codes[fillInPos].vmcode = offset;
  }
}


//------------------------------------
  void BackEnd::fillInBranchOffsets ()
//------------------------------------
{
  std::map<unsigned short, unsigned>::iterator iter;

  for ( iter = jumps.begin(); iter != jumps.end(); ++iter )
  {
    unsigned short fillInPos = (*iter).first;
    unsigned mark = (*iter).second;
    unsigned short offset = (unsigned short) (marks[mark] - fillInPos - 1);
    codes[fillInPos].vmcode = offset;
  }
}



//--- protected code putting routines -------------------------------

//---------------------------------------------
  void BackEnd::putCode ( unsigned short code )
//---------------------------------------------
{
  codes[codeOffset].vmcode = code;
  codes[codeOffset].instructionNo = instructionNo;
  codeOffset++;
}


//----------------------------------------------------------
  void BackEnd::putBranchOffset ( IntermediateCode& imcode )
//----------------------------------------------------------
{
  jumps[(unsigned short)codeOffset] = imcode.getID();
  putCode(0); 
}


//---------------------------------------------------
  void BackEnd::putCalcCode_A ( unsigned short code )
//---------------------------------------------------
{
  int calcTypeID = getCalcTypeID(); 
  putCode( (unsigned short)(code | (calcTypeID << 8)) );

  switch ( calcTypeID )
  {
    case CALC_INT_INT     : types.push(NumericType::INT); break;
    case CALC_INT_LONG    : types.push(NumericType::LONG); break;
    case CALC_INT_FLOAT   : types.push(NumericType::FLOAT); break;
    case CALC_LONG_INT    : types.push(NumericType::LONG); break;
    case CALC_LONG_LONG   : types.push(NumericType::LONG); break;
    case CALC_LONG_FLOAT  : types.push(NumericType::FLOAT); break;
    case CALC_FLOAT_INT   : types.push(NumericType::FLOAT); break;
    case CALC_FLOAT_LONG  : types.push(NumericType::FLOAT); break;
    case CALC_FLOAT_FLOAT : types.push(NumericType::FLOAT); break;

    default : break;
  }
}


//---------------------------------------------------
  void BackEnd::putCalcCode_B ( unsigned short code )
//---------------------------------------------------
{
  int calcTypeID = getCalcTypeID(); 
  putCode( (unsigned short)(code | (calcTypeID << 8)) );

  types.push(NumericType::INT);
}


//---------------------------------------------------
  void BackEnd::putCalcCode_C ( unsigned short code )
//---------------------------------------------------
{
  int calcTypeID = getCalcTypeID(); 
  putCode( (unsigned short)(code | (calcTypeID << 8)) );

  switch ( calcTypeID )
  {
    case CALC_INT_INT     : types.push(NumericType::INT); break;
    case CALC_INT_LONG    : types.push(NumericType::LONG); break;
    case CALC_INT_FLOAT   : types.push(NumericType::INT); break;
    case CALC_LONG_INT    : types.push(NumericType::LONG); break;
    case CALC_LONG_LONG   : types.push(NumericType::LONG); break;
    case CALC_LONG_FLOAT  : types.push(NumericType::INT); break;
    case CALC_FLOAT_INT   : types.push(NumericType::INT); break;
    case CALC_FLOAT_LONG  : types.push(NumericType::INT); break;
    case CALC_FLOAT_FLOAT : types.push(NumericType::INT); break;

    default : break;
  }
}


//--------------------------------------------------------
  void BackEnd::put_LOAD_code ( IntermediateCode& imcode )
//--------------------------------------------------------
{
  if ( typeid(*imcode.getObject()) == typeid(Variable) )
  {
    Variable* variable = dynamic_cast<Variable*>(imcode.getObject());
    NumericType* type = dynamic_cast<NumericType*>(variable->getType());
    NumericType::FORM typeID = type->getForm();
  
    if ( variable->isIndirectVariable() )
    {
      putCode(VM_LOAD_LOCAL_INT);
      putCode((unsigned short)(variable->getReferenceOffset() - 4));
      putCode(VM_LOAD_IMMEDIATE_INT);
      putCode((unsigned short)variable->getOffset());
      putCode((unsigned short)(VM_ADD | (CALC_INT_INT << 8)));

      switch ( typeID )
      {
        case NumericType::BYTE  : putCode(VM_LOAD_REF_BYTE); typeID = NumericType::INT; break;
        case NumericType::INT   : putCode(VM_LOAD_REF_INT); break;
        case NumericType::LONG  : putCode(VM_LOAD_REF_LONG); break;
        case NumericType::FLOAT : putCode(VM_LOAD_REF_FLOAT); break;
      }
    } 
    else if ( variable->getMemoryClass() == Variable::STATIC )
    {
      switch ( typeID )
      {
        case NumericType::BYTE  : putCode(VM_LOAD_BYTE); typeID = NumericType::INT; break;
        case NumericType::INT   : putCode(VM_LOAD_INT); break;
        case NumericType::LONG  : putCode(VM_LOAD_LONG); break;
        case NumericType::FLOAT : putCode(VM_LOAD_FLOAT); break;
      }
//xxx ??      putCode((unsigned short)(varBase + variable->getOffset()));
      putCode((unsigned short)variable->getOffset());
    }
    else
    {
      switch ( typeID )
      {
        case NumericType::BYTE  : putCode(VM_LOAD_LOCAL_BYTE); typeID = NumericType::INT; break;
        case NumericType::INT   : putCode(VM_LOAD_LOCAL_INT); break;
        case NumericType::LONG  : putCode(VM_LOAD_LOCAL_LONG); break;
        case NumericType::FLOAT : putCode(VM_LOAD_LOCAL_FLOAT); break;
      }
      putCode((unsigned short)(variable->getOffset() - ( variable->getOffset() < 0 ? 4 : 0)));  // bp and pc on stack!
    }
    types.push(typeID);
  }
  else
  {
    Constant* constant = dynamic_cast<Constant*>(imcode.getObject());
    NumericType* type = dynamic_cast<NumericType*>(constant->getType());
    NumericType::FORM typeID = type->getForm();

    switch ( typeID )
    {
      case NumericType::BYTE  : putCode(VM_LOAD_CONST_BYTE); typeID = NumericType::INT; break;
      case NumericType::INT   : putCode(VM_LOAD_CONST_INT); break;
      case NumericType::LONG  : putCode(VM_LOAD_CONST_LONG); break;
      case NumericType::FLOAT : putCode(VM_LOAD_CONST_FLOAT); break;
    }
    putCode((unsigned short)constant->getOffset()); 
    types.push(typeID);
  }
}


//----------------------------------------------------------
  void BackEnd::put_R_LOAD_code ( IntermediateCode& imcode )
//----------------------------------------------------------
{
  bool increment = (imcode.opcode == IntermediateCode::R_INC_LOAD);

  put_CAST_code (NumericType::INT);
  
  Data* data = dynamic_cast<Data*>(imcode.getObject());
  NumericType* type = dynamic_cast<NumericType*>(data->getType());
  NumericType::FORM typeID = type->getForm();

  if ( data->getMemoryClass() == Data::CONSTANT )
  {
    Constant* constant = dynamic_cast<Constant*>(data);
  
    switch ( typeID )
    {
      case NumericType::BYTE  : putCode(VM_LOAD_REF_CONST_BYTE); typeID = NumericType::INT; break;
      case NumericType::INT   : putCode(VM_LOAD_REF_CONST_INT); break;
      case NumericType::LONG  : putCode(VM_LOAD_REF_CONST_LONG); break;
      case NumericType::FLOAT : putCode(VM_LOAD_REF_CONST_FLOAT); break;
    }
  }
  else  
  {
    Variable* variable = dynamic_cast<Variable*>(data);

    switch ( typeID )
    {
      case NumericType::BYTE  : putCode(increment? VM_INC_LOAD_REF_BYTE : VM_LOAD_REF_BYTE); typeID = NumericType::INT; break;
      case NumericType::INT   : putCode(increment? VM_INC_LOAD_REF_INT : VM_LOAD_REF_INT); break;
      case NumericType::LONG  : putCode(increment? VM_INC_LOAD_REF_LONG : VM_LOAD_REF_LONG); break;
      case NumericType::FLOAT : putCode(increment? VM_INC_LOAD_REF_FLOAT : VM_LOAD_REF_FLOAT); break;
    }
  }

  types.pop();
  types.push(typeID);
}


//-------------------------------------------
  void BackEnd::put_ILOAD_code ( long value )
//-------------------------------------------
{
  if ( (value >= 0) && (value <= 255) )
  {
    putCode((unsigned short)((value << 8) | VM_LOAD_IMMEDIATE_BYTE));
    types.push(NumericType::INT); // !!!
  }
  else if ( (value >= -32768) && (value <= 32767) )
  {
    putCode(VM_LOAD_IMMEDIATE_INT);
    putCode((unsigned short)value);
    types.push(NumericType::INT);
  }
  else
  {
    putCode(VM_LOAD_CONST_LONG);
    putCode(storeConst(value));
    types.push(NumericType::LONG);
  }
}


//-----------------------------------------------
  void BackEnd::put_ILOAD_F_code ( double value )
//-----------------------------------------------
{
  short ivalue = (short) value;

  NumericType::FORM typeID = StandardType::getFromValue(value)->getForm();

  switch ( typeID )
  {
    case NumericType::BYTE :

      putCode( (unsigned short)((ivalue << 8) | VM_LOAD_IMMEDIATE_BYTE) );
      types.push(NumericType::INT); // !!!
      break;
    
    case NumericType::INT : 

      putCode(VM_LOAD_IMMEDIATE_INT);
      putCode( (unsigned short) ivalue );
      types.push(NumericType::INT);
      break;

    case NumericType::LONG : 
    
      putCode(VM_LOAD_CONST_LONG);
      putCode(storeConst((long) value));
      types.push(NumericType::LONG);
      break;

    case NumericType::FLOAT : 

      putCode(VM_LOAD_CONST_FLOAT);
      putCode(storeConst((double) value));
      types.push(NumericType::FLOAT);
      break;
  }
}


//------------------------------------------------------------
  void BackEnd::put_INC_LOAD_code ( IntermediateCode& imcode )
//------------------------------------------------------------
{
  Variable* variable = dynamic_cast<Variable*>(imcode.getObject());
  NumericType* type = dynamic_cast<NumericType*>(variable->getType());
  NumericType::FORM typeID = type->getForm();

  if ( variable->isIndirectVariable() )
  {
    putCode(VM_LOAD_LOCAL_INT);
    putCode((unsigned short)(variable->getReferenceOffset() - 4));
    putCode(VM_LOAD_IMMEDIATE_INT);
    putCode((unsigned short)variable->getOffset());
    putCode((unsigned short)(VM_ADD | (CALC_INT_INT << 8)));

    switch ( typeID )
    {
      case NumericType::BYTE  : putCode(VM_INC_LOAD_REF_BYTE); typeID = NumericType::INT; break;
      case NumericType::INT   : putCode(VM_INC_LOAD_REF_INT); break;
      case NumericType::LONG  : putCode(VM_INC_LOAD_REF_LONG); break;
      case NumericType::FLOAT : putCode(VM_INC_LOAD_REF_FLOAT); break;
    }
    putCode((unsigned short)(variable->getOffset() - ( variable->getOffset() < 0 ? 4 : 0)));  // bp and pc on stack!
  }
  else if ( variable->getMemoryClass() == Variable::STATIC )
  {
    switch ( typeID )
    {
      case NumericType::BYTE  : putCode(VM_INC_LOAD_BYTE); typeID = NumericType::INT; break;
      case NumericType::INT   : putCode(VM_INC_LOAD_INT); break;
      case NumericType::LONG  : putCode(VM_INC_LOAD_LONG); break;
      case NumericType::FLOAT : putCode(VM_INC_LOAD_FLOAT); break;
    }
// xxx??    putCode((unsigned short)(varBase + variable->getOffset()));
    putCode((unsigned short)variable->getOffset());
  }
  else
  {
    switch ( typeID )
    {
      case NumericType::BYTE  : putCode(VM_INC_LOAD_LOCAL_BYTE); typeID = NumericType::INT; break;
      case NumericType::INT   : putCode(VM_INC_LOAD_LOCAL_INT); break;
      case NumericType::LONG  : putCode(VM_INC_LOAD_LOCAL_LONG); break;
      case NumericType::FLOAT : putCode(VM_INC_LOAD_LOCAL_FLOAT); break;
    }
    putCode((unsigned short)(variable->getOffset() - ( variable->getOffset() < 0 ? 4 : 0)));  // bp and pc on stack!
  }

  types.push(typeID);
}


//---------------------------------------------------------
  void BackEnd::put_STORE_code ( IntermediateCode& imcode )
//---------------------------------------------------------
{
  Variable* variable = dynamic_cast<Variable*>(imcode.getObject());
  NumericType* type = dynamic_cast<NumericType*>(variable->getType());
  NumericType::FORM typeID = type->getForm();

  bool keep = (imcode.opcode == IntermediateCode::K_STORE);

  put_CAST_code(typeID);

  if ( variable->isIndirectVariable() )
  {
    putCode(VM_LOAD_LOCAL_INT);
    putCode((unsigned short)(variable->getReferenceOffset() - 4));
    putCode(VM_LOAD_IMMEDIATE_INT);
    putCode((unsigned short)variable->getOffset());
    putCode(VM_ADD | (CALC_INT_INT << 8));

    switch ( typeID )
    {
      case NumericType::BYTE  : putCode(VM_STORE_REF_BYTE); break;
      case NumericType::INT   : putCode(VM_STORE_REF_INT); break;
      case NumericType::LONG  : putCode(VM_STORE_REF_LONG); break;
      case NumericType::FLOAT : putCode(VM_STORE_REF_FLOAT); break;
    }
  }
  else if ( variable->getMemoryClass() == Variable::STATIC )
  {
    switch ( typeID )
    {
      case NumericType::BYTE  : putCode(keep? VM_STORE_KEEP_BYTE : VM_STORE_BYTE); break;
      case NumericType::INT   : putCode(keep? VM_STORE_KEEP_INT : VM_STORE_INT); break;
      case NumericType::LONG  : putCode(keep? VM_STORE_KEEP_LONG : VM_STORE_LONG); break;
      case NumericType::FLOAT : putCode(keep? VM_STORE_KEEP_FLOAT : VM_STORE_FLOAT); break;
    }
//xxx??    putCode((unsigned short)(varBase + variable->getOffset()));
      putCode((unsigned short)variable->getOffset());
  }
  else
  {
    switch ( typeID )
    {
      case NumericType::BYTE  : putCode(keep? VM_STORE_KEEP_LOCAL_BYTE : VM_STORE_LOCAL_BYTE); break;
      case NumericType::INT   : putCode(keep? VM_STORE_KEEP_LOCAL_INT : VM_STORE_LOCAL_INT); break;
      case NumericType::LONG  : putCode(keep? VM_STORE_KEEP_LOCAL_LONG : VM_STORE_LOCAL_LONG); break;
      case NumericType::FLOAT : putCode(keep? VM_STORE_KEEP_LOCAL_FLOAT : VM_STORE_LOCAL_FLOAT); break;
    }
    putCode((unsigned short)(variable->getOffset() - ( variable->getOffset() < 0 ? 4 : 0)));  // bp and pc on stack!
  }

  if ( !keep )
    types.pop();
}


//-----------------------------------------------------------
  void BackEnd::put_R_STORE_code ( IntermediateCode& imcode )
//-----------------------------------------------------------
{
  Variable* variable = dynamic_cast<Variable*>(imcode.getObject());
  NumericType* type = dynamic_cast<NumericType*>(variable->getType());
  NumericType::FORM typeID = type->getForm();

  switch ( typeID )
  {
    case NumericType::BYTE  : putCode(VM_STORE_REF_BYTE); break;
    case NumericType::INT   : putCode(VM_STORE_REF_INT); break;
    case NumericType::LONG  : putCode(VM_STORE_REF_LONG); break;
    case NumericType::FLOAT : putCode(VM_STORE_REF_FLOAT); break;
  }

  types.pop();
  types.pop(); 
}


//--------------------------------------------------------
  void BackEnd::put_CALL_code ( IntermediateCode& imcode )
//--------------------------------------------------------
{
  Function* function = dynamic_cast<Function*>(imcode.getObject());
  Type* type = function->getType();

  if ( !function->isInline() )
  {
    std::string calledFunctionName;
    getFullName(function->getIdentifier(), calledFunctionName);
    putCode(VM_CALL);
    calls[(unsigned short)codeOffset] = calledFunctionName;
    putCode(0);
  }

  for ( int i = 0; i < function->getNumParams(); i++ )
    types.pop();

  if ( type )
  {
    NumericType::FORM typeID = (dynamic_cast<NumericType*>(type))->getForm();
    types.push(typeID);
  }
}


//----------------------------------------------------------
  void BackEnd::put_RETURN_code ( IntermediateCode& imcode )
//----------------------------------------------------------
{
  Function* function = dynamic_cast<Function*>(imcode.getObject());

  if ( !function->getType() )
  {
    if ( function->getNumParams() || function->getImCodeSeq()->getLocalMemSize() )
    {
      putCode(VM_RETURN); 
      putCode((unsigned short)(function->getParamMemSize() + 4));  // 4 because of bp and pc on stack with 2 bytes each
    }
    else
      putCode(VM_FAST_RETURN);
  }
  else 
  {
    NumericType* type = dynamic_cast<NumericType*>(function->getType());

    switch ( type->getForm() )
    {
      case NumericType::BYTE : 
      case NumericType::INT : 

         put_CAST_code(NumericType::INT);
         putCode(VM_RETURN_VALUE | (RTYPE_INT << 8)); 
         break;

      case NumericType::LONG : 

         put_CAST_code(NumericType::LONG);
         putCode(VM_RETURN_VALUE | (RTYPE_LONG << 8)); 
         break;

      case NumericType::FLOAT : 

         put_CAST_code(NumericType::FLOAT);
         putCode(VM_RETURN_VALUE | (RTYPE_FLOAT << 8)); 
         break;
    }
    putCode((unsigned short)(function->getParamMemSize() + 4));  // 4 because of bp and pc on stack with 2 bytes each
  }
}


//-------------------------------------------------------------
  void BackEnd::put_WIPESTACK_code ( IntermediateCode& imcode )
//-------------------------------------------------------------
{
  Function* function = dynamic_cast<Function*>(imcode.getObject());
  NumericType* type = dynamic_cast<NumericType*>(function->getType());

  putCode(VM_INC_SP); 

  switch ( type->getForm() )
  {
    case NumericType::BYTE  : 
    case NumericType::INT   : putCode(-2); break;
    case NumericType::LONG  : putCode(-4); break;
    case NumericType::FLOAT : putCode(-8); break;
  }
}



//------------------------------------------------------------
  void BackEnd::put_LOAD_PTR_code ( IntermediateCode& imcode )
//------------------------------------------------------------
{
  Data* data = dynamic_cast<Data*>(imcode.getObject());

  if ( data->getMemoryClass() == Data::CONSTANT )
  {
    putCode(VM_LOAD_IMMEDIATE_INT);
    putCode((unsigned short)data->getOffset());
  }
  else
  {
    Variable* variable = dynamic_cast<Variable*>(imcode.getObject());

    if ( variable->isIndirectVariable() )
    {
      putCode(VM_LOAD_LOCAL_INT);
      putCode((unsigned short)(variable->getReferenceOffset() - 4));
    }
    else 
    {
      putCode(VM_LOAD_IMMEDIATE_INT);
      putCode((unsigned short)variable->getOffset());

      if ( variable->getMemoryClass() == Data::LOCAL )
        putCode(VM_MAKE_GLOBAL_POINTER);
    }
  }

  types.push(NumericType::INT);
}


//--------------------------------------------------------
  void BackEnd::put_CAST_code ( IntermediateCode& imcode )
//--------------------------------------------------------
{
  put_CAST_code((NumericType::FORM)imcode.getID());
}


//--------------------------------------------------------
  void BackEnd::put_CAST_code ( NumericType::FORM typeID )
//--------------------------------------------------------
{
  NumericType::FORM typeOnStackID = types.top(); 
  types.pop();

  switch ( typeID )
  {
    case NumericType::BYTE : 
    case NumericType::INT :
    
      switch ( typeOnStackID )
      {
        case NumericType::LONG  : putCode(VM_CAST_LONG_TO_INT); break;
        case NumericType::FLOAT : putCode(VM_CAST_FLOAT_TO_INT); break;

        default : break;
      }
      break;

    case NumericType::LONG : 
    
      switch ( typeOnStackID )
      {
        case NumericType::INT   : putCode(VM_CAST_INT_TO_LONG); break;
        case NumericType::FLOAT : putCode(VM_CAST_FLOAT_TO_LONG); break;

        default : break;
      }
      break;

    case NumericType::FLOAT : 
    
      switch ( typeOnStackID )
      {
        case NumericType::INT  : putCode(VM_CAST_INT_TO_FLOAT); break;
        case NumericType::LONG : putCode(VM_CAST_LONG_TO_FLOAT); break;

        default : break;
      }
      break;

    default : break;
  }

  types.push(typeID);
}

 
//-----------------------------
  void BackEnd::put_NEG_code ()
//-----------------------------
{
  switch ( types.top() )
  {
    case NumericType::BYTE  : 
    case NumericType::INT   : putCode(VM_NEG_INT); break;
    case NumericType::LONG  : putCode(VM_NEG_LONG); break;
    case NumericType::FLOAT : putCode(VM_NEG_FLOAT); break;
  }
}


//-----------------------------
  void BackEnd::put_NOT_code ()
//-----------------------------
{
  switch ( types.top() )
  {
    case NumericType::BYTE  : 
    case NumericType::INT   : putCode(VM_INVERT_INT); break;
    case NumericType::LONG  : putCode(VM_INVERT_LONG); break;

    case NumericType::FLOAT : 

      putCode(VM_INVERT_FLOAT); 
      types.pop();
      types.push(NumericType::INT);
      break;
  }
}
 

//--- string operations backend -------------------------------------


//----------------------------------------------------------
  void BackEnd::put_STR_PUTISTR_code ( std::string& string )
//----------------------------------------------------------
{
  unsigned short offset = constOffset;
  const int MAX_LENGTH = 30;
  int length = string.length() < MAX_LENGTH ? string.length() : MAX_LENGTH;
  
  unsigned char* ptr = (unsigned char*) (constants + offset);
  ptr[0] = (unsigned char)length;

  int i;

  for ( i = 1; i <= length; i++ )
    ptr[i] = (unsigned char) string[i-1]; 

  constOffset += (i + 1) / 2;

  putCode(VM_LOAD_IMMEDIATE_INT);
  putCode(offset);
  putCode( VM_STRING | (STRING_PUT_CONST_STRING << 8));
}


//---------------------------------------------------------------
  void BackEnd::put_STR_PUTCSTR_code ( IntermediateCode& imcode )
//---------------------------------------------------------------
{
  Data* data = dynamic_cast<Data*>(imcode.getObject());

  if ( data )
  {
    if ( data->isArray() )
    {
      putCode(VM_LOAD_REF_CONST_INT);
    }
    else
    { 
      StringConstant* strconst = dynamic_cast<StringConstant*>(data);
      putCode(VM_LOAD_IMMEDIATE_INT);
      putCode(strconst->getOffset());
    }
  }

  putCode( VM_STRING | (STRING_PUT_CONST_STRING << 8));  
}




