/********************************************************************

    project C-Control II Compiler

    branch.h

********************************************************************/

#ifndef branch_h
#define branch_h

#include "stl.h"


//--------------------------------
  template<class Key, class Value> 
  class Branch
//--------------------------------
{
  friend class Branch;

  protected:

    typedef std::list<Branch*> List;

    Key     key;
    Value   value;
    Branch* parent;
    List    children;
    bool    hiddenChildren;

    List::iterator childIterator;

    Branch* getChildX ( const Key& _key );

    Branch ();

  public:
        
    Branch ( const Key& _key, const Value _value );
   ~Branch ();

    const Key&  getKey ();
    const Value& getValue ();

    int  getNumChildren ();
    void hideChildren ( bool hide = true );

    Branch* addChild ( const Key& _key, const Value _value );
    Branch* getChild ( const Key& _key );
    Branch* getFirstChild ();
    Branch* getNextChild ();
    Branch* seekChild ( const Key& _key );

    Branch* getParent ();

    void removeChildren ();
    bool removeChild ( const Key& _key );
    bool remove (); 

    void writePath ( std::ostream& os, char separator );
};


template<class Key, class Value>
Branch<Key, Value>::Branch ( const Key& _key, const Value _value )
{ 
  key = _key; 
  value = _value; 
  parent = 0; 
  hiddenChildren = false; 
}


template<class Key, class Value>
Branch<Key, Value>::~Branch ()
{ 
  removeChildren();
}



template<class Key, class Value>
inline const Key& Branch<Key, Value>::getKey ()
{ return key; }


template<class Key, class Value>
inline const Value& Branch<Key, Value>::getValue ()
{ return value; }


template<class Key, class Value>
inline void Branch<Key, Value>::hideChildren ( bool hide )
{ hiddenChildren = hide; }


template<class Key, class Value>
inline int Branch<Key, Value>::getNumChildren ()
{ return children.size(); }


template<class Key, class Value>
Branch<Key, Value>* Branch<Key, Value>::addChild ( const Key& _key, const Value _value )
{
  if ( getChildX(_key) )
    return 0;

  Branch<Key, Value>* child = new Branch<Key, Value>(_key, _value);

  children.push_back(child);
  child->parent = this;
  return child;
}


template<class Key, class Value>
Branch<Key, Value>* Branch<Key, Value>::getChild ( const Key& _key )
{
  if ( ! hiddenChildren )
    return getChildX(_key);

  return 0;
}


template<class Key, class Value>
Branch<Key, Value>* Branch<Key, Value>::getChildX ( const Key& _key )
{
  for ( List::iterator iterator = children.begin(); iterator != children.end(); ++iterator )
    if ( (*iterator)->getKey() == _key )
      return (*iterator);

  return 0;
}


template<class Key, class Value>
Branch<Key, Value>* Branch<Key, Value>::getFirstChild ()
{
  childIterator = children.begin();
  return (*childIterator);
}


template<class Key, class Value>
Branch<Key, Value>* Branch<Key, Value>::getNextChild ()
{
  ++childIterator;
  return (*childIterator);
}


template<class Key, class Value>
Branch<Key, Value>* Branch<Key, Value>::seekChild ( const Key& _key )
{
  for ( Branch<Key, Value>* branch = this; branch; branch = branch->parent )
  {
    Branch<Key, Value>* child = (branch == this)? branch->getChildX(_key) : branch->getChild(_key);

    if ( child )
      return child;
  }

  return 0;
}


template<class Key, class Value>
Branch<Key, Value>* Branch<Key, Value>::getParent ()
{ return parent; }


template<class Key, class Value>
void Branch<Key, Value>::removeChildren ()
{ 
  for ( List::iterator iterator = children.begin(); iterator != children.end(); ++iterator )
    delete (*iterator);

  children.clear(); 
}


template<class Key, class Value>
bool Branch<Key, Value>::removeChild ( const Key& _key )
{
  for ( List::iterator iterator = children.begin(); iterator != children.end(); ++iterator )
    if ( (*iterator)->getKey() == _key )
    {
      delete (*iterator);
      children.erase(iterator);
      return true;
    }

  return false;
}


template<class Key, class Value>
bool Branch<Key, Value>::remove ()
{
  if ( parent )
    return parent->removeChild(key);
  else
    return false;
}


template<class Key, class Value>
void Branch<Key, Value>::writePath ( std::ostream& os, char separator )
{
  std::ostrstream ss;
  Branch* branch;
  std::list<Branch*> path;
  std::list<Branch*>::iterator iterator;

  for ( branch = this; branch; branch = branch->parent )
    path.push_front(branch);
 
  for ( iterator = path.begin(); iterator != path.end(); ++iterator )
  {
    if ( ss.pcount() )
      ss << separator; 

    ss << (*iterator)->getKey();
  }

  ss << '\x0';
  os << ss.str();
  ss.freeze(0);
}


#endif