/********************************************************************

    project C-Control II Compiler

    imcode.h

********************************************************************/

#ifndef imcode_h
#define imcode_h

#include "classlib/classlib.h"


//-----------------------
  struct IntermediateCode
//-----------------------
{
  typedef enum 
  { 
    NOP, CALL, RETURN, WIPESTACK, 
    MARK, GOTO, BRANCH_IF_ZERO, 
    CAST,                
    LOAD, I_LOAD, I_LOAD_F, LOAD_PTR, R_LOAD, INC_LOAD, R_INC_LOAD, STORE, R_STORE, K_STORE,
    NEG, NOT,
    ADD, SUB, MUL, DIV, MOD, 
    EQUAL, NOTEQUAL, HIGHER, LOWER, HIGHERSAME, LOWERSAME, 
    AND, NAND, OR, NOR, XOR, NXOR, 
    SHL, SHR, ASR, ROL, ROR, 

    STR_CLEAR, STR_PUTCH, STR_PUTSTR, STR_PUTISTR, STR_PUTCSTR, STR_PUTVAL,
                    
    INLINE, RUN, HALT, RESUME, RESET, 
    SLEEP, CAPTURE, RELEASE, YIELD, QUIT,

    ENTER_INSTRUCTION
                  
  } OPERATION;

  OPERATION opcode; 
  
  union 
  {  
    long     value;     
    Virtual* object;
  } enclosed;

  void setValue ( long value );
  void setObject ( Virtual* object );

  long     getValue () const;
  Virtual* getObject () const;
  unsigned getID () const;

  static bool isNumericalOperation ( OPERATION opcode );
  bool hasConstantValue () const;
};

inline void IntermediateCode::setValue ( long value )
{ enclosed.value = value; }

inline void IntermediateCode::setObject ( Virtual* object )
{ enclosed.object = object; }

inline long IntermediateCode::getValue () const
{ return enclosed.value; }

inline Virtual* IntermediateCode::getObject () const
{ return enclosed.object; }

inline unsigned IntermediateCode::getID () const
{ return (unsigned) getValue(); }

inline bool IntermediateCode::isNumericalOperation ( OPERATION opcode )
{ return (opcode >= NEG) && (opcode <= ROR); }




//------------------------------
  class IntermediateCodeSequence 
//------------------------------
{
  protected:

    IntermediateCode* codes;
    double*           floats;
    std::string*      strings;

    unsigned long codeCapacity;
    unsigned long floatCapacity;
    unsigned long stringCapacity;

    unsigned long numberOfCodes;
    unsigned long numberOfFloats;
    unsigned long numberOfStrings;

    unsigned long iCode;
    unsigned long iFloat;
    unsigned long iString;

    //-----------------------------


    static unsigned nextmark;
    
    unsigned ownMemoryRequirements;
    unsigned callMemoryRequirements;
    unsigned localMemSize;

    unsigned constCounter;
    
  public:

    IntermediateCodeSequence ();
   ~IntermediateCodeSequence ();

    void init ();
    void free ();
    void finish ();

    void rewind ();
    void gotoNextCode ();
    bool isLastCode () const;
    
    IntermediateCode& codesBack () const;
    IntermediateCode& getCode () const;
    std::string& getCodeEnclosedString () const;
    double getCodeEnclosedFloat () const;

    void putCode ( IntermediateCode::OPERATION opcode ); 
    void putCodeWithID ( IntermediateCode::OPERATION opcode, unsigned id );
    void putImmLoadValueCode ( long value );
    void putImmLoadFloatValueCode ( double value );
    void putCodeWithObject ( IntermediateCode::OPERATION opcode, Virtual* object );
    void putCodeWithString ( IntermediateCode::OPERATION opcode, const std::string& string );
 
    void pushCode ( IntermediateCode::OPERATION opcode ); 
    void pushCode ( IntermediateCode::OPERATION opcode, long value ); 
    void pushCode ( IntermediateCode::OPERATION opcode, Virtual* object ); 
    void pushFloatCode ( IntermediateCode::OPERATION opcode, double value );
    void pushStringCode ( IntermediateCode::OPERATION opcode, const std::string& string );
    void popCode ();

    void chainCodeSeq ( IntermediateCodeSequence* _seq, bool shallCopyInstructionChangeInfo );
    bool isEmpty () const;

    static unsigned getMark ();
    unsigned mark ();
    void mark ( unsigned id );

    void   beginConstantFolding (); 
    double getConstantResult ();
    class NO_CONST {};

    void dump ( std::ostream& os );

    void requireMemory ( unsigned size );
    void requireMemoryForCall ( unsigned size );

    void setLocalMemSize ( unsigned size );
    unsigned getLocalMemSize ();
    unsigned getMemoryRequirements ();
};


inline void IntermediateCodeSequence::rewind ()
{ iCode = 0; }

inline void IntermediateCodeSequence::gotoNextCode ()
{ ++iCode; }

inline bool IntermediateCodeSequence::isLastCode () const
{ return iCode == numberOfCodes; }

inline bool IntermediateCodeSequence::isEmpty () const
{ return !iCode; }

inline IntermediateCode& IntermediateCodeSequence::codesBack () const
{ return codes[iCode-1]; }

inline IntermediateCode& IntermediateCodeSequence::getCode () const
{ return codes[iCode]; }

inline std::string& IntermediateCodeSequence::getCodeEnclosedString () const
{ return strings[getCode().getValue()]; }

inline double IntermediateCodeSequence::getCodeEnclosedFloat () const
{ return floats[getCode().getValue()]; }

inline void IntermediateCodeSequence::popCode ()
{ --iCode; }

inline void IntermediateCodeSequence::setLocalMemSize ( unsigned size )
{ localMemSize = size; } 

inline unsigned IntermediateCodeSequence::getLocalMemSize ()
{ return localMemSize; }

inline void IntermediateCodeSequence::beginConstantFolding ()
{ constCounter = 0; }


#endif