/********************************************************************

    project C-Control II Compiler

    parsdef.cpp

********************************************************************/


#include "parser.h"
#include "error.h"
#include "text.h"


//------------------------------------------------------------
  unsigned Parser::parseNameWithBrackets ( std::string& name )
//------------------------------------------------------------
{
  unsigned number = 0;

  readNextToken(Token::IDENTIFIER);
  name = input->getString();
 
  if ( readNextToken() == Token::LEFT_BRACKET )
  {   
    number = parseConstTerm();
    readNextToken(Token::RIGHT_BRACKET);
    readNextToken();
  }

  return number;
}


//--------------------------------
  void Parser::parseUserTypeDef ()
//--------------------------------
{
  declarationLineNo = lineNo;
  readNextToken(Token::IDENTIFIER); 

  UserType*   userType   = new UserType();
  Identifier* identifier = declare(input->getString(), userType, TYPE_IDENT);

  program->enterScope(identifier);

  for ( readNextToken(Token::LEFT_BRACE); readNextToken() != Token::RIGHT_BRACE; )  
  {
    if ( input->getToken().id == Token::SEMICOLON )
      continue;

    Type* type = parseType();
    std::string name;
    declarationLineNo = lineNo;
    unsigned numArrayElements = parseNameWithBrackets(name);
    UserType::Field* field    = userType->newField(type, numArrayElements);
    Identifier* identifier    = declare(name, field, FIELD_IDENT);

    if ( field->getType() == userType ) 
    {
      std::string errMsg;
      errMsg = Text::get(Text::RECURSIVE_TYPE);
      complain(errMsg);
      throw Error();
    }

    expect(Token::SEMICOLON);
  }

  program->leaveScope();

}


//--------------------------
  Type* Parser::parseType ()
//--------------------------
{
  Identifier* identifier;
  Type* type;

  switch ( input->getToken().id )
  {
    case Token::BYTE   : type = &StandardType::BYTE; break;
    case Token::INT    : type = &StandardType::INT; break;
    case Token::LONG   : type = &StandardType::LONG; break;
    case Token::FLOAT  : type = &StandardType::FLOAT; break;
    case Token::STRING : type = &StandardType::STRING; break;

    case Token::IDENTIFIER:

      identifier = parseIdentifier();
      expect(identifier, TYPE_IDENT);
      type = dynamic_cast<Type*>(identifier->getValue().get());
      break;  

    default: unexpect(input->getToken().id);
  }

  return type;
}


//-----------------------------
  void Parser::parseConstDef ()
//-----------------------------
{
  bool isArray = false;
  
  declarationLineNo = lineNo;
  readNextToken(Token::IDENTIFIER);
  std::string name = input->getString();
 
  if ( readNextToken() == Token::LEFT_BRACKET )
  {   
    readNextToken(Token::RIGHT_BRACKET);
    isArray = true;
  }
  else
    putBackToken();

  readNextToken(Token::ASSIGN);
  readNextToken();

  if ( input->getToken().id == Token::LITERAL_STRING )
  {
    if ( isArray )
    {
      StringArrayConstant* c = new StringArrayConstant();
      putBackToken();

      do
      { 
        readNextToken(Token::LITERAL_STRING);
        c->addString(input->getString()); 

      } while ( readNextToken() == Token::COMMA );

      putBackToken();     
      declare(name, c, STRING_ARRAY_CONST_IDENT);
    }
    else
    {
      StringConstant* c = new StringConstant();
      c->setString(input->getString());
      declare(name, c, STRING_CONST_IDENT);
    }
  }
  else
  {
    if ( isArray )
    {
      ArrayConstant* c = new ArrayConstant();
      putBackToken();

      do
      {
        if ( (c->getNumArrayElements() == 65535) || (c->getSize() > 130000) )
        {
          std::string errMsg;
          errMsg = Text::get(Text::ARRAY_TOO_BIG);
          complain(errMsg);
          throw Error();
        }

        c->addValue(parseConstTerm()); 
      }
      while ( readNextToken() == Token::COMMA );

      putBackToken();     
      declare(name, c, ARRAY_CONST_IDENT);
    }
    else
    {
      Constant* c = new Constant();
      putBackToken();
      c->setValue(parseConstTerm());
      declare(name, c, CONST_IDENT);
    }
  }
}


//---------------------------
  void Parser::parseVarDef ()
//---------------------------
{
  Type* type = parseType();

  do 
  { 
    std::string name;
    declarationLineNo = lineNo;
    unsigned long numElements = parseNameWithBrackets(name);

    if ( numElements > 65535 )
    {
      std::string errMsg;
      errMsg = Text::get(Text::ARRAY_TOO_BIG);
      complain(errMsg);
      throw Error();
    }

    Variable* variable = new Variable( type, isLocal()? Data::LOCAL : Data::STATIC, numElements != 0, numElements);   
    declare(name, variable, VAR_IDENT);

    if ( isLocal() )
      currentlyParsedProc->addLocalVariable(variable);

    if ( type->isUserType() )  
      declareUserTypeFieldVars(variable);

  }	while ( input->getToken().id == Token::COMMA );

  putBackToken();
}


//------------------------------------------------------------
  void Parser::declareUserTypeFieldVars ( Variable* variable )
//------------------------------------------------------------
{
  UserType* usertype = dynamic_cast<UserType*>(variable->getType());
  Identifier* userTypeIdent = usertype->getIdentifier();
  unsigned numFields = userTypeIdent->getNumChildren();
  Identifier* fieldIdent;
  unsigned i;

  program->enterScope(variable->getIdentifier());

  for ( i = 0, fieldIdent = userTypeIdent->getFirstChild(); 
       i < numFields; 
       i++, fieldIdent = userTypeIdent->getNextChild() )
  { 
    UserType::Field* field = dynamic_cast<UserType::Field*>(fieldIdent->getValue().get());

    Variable* fieldVar = new Variable( field->getType(), variable->getMemoryClass(), 
                                       field->isArray(), field->getNumArrayElements(), 
                                       variable->getReference());   
     
    declare(fieldIdent->getKey(), fieldVar, VAR_IDENT);
    fieldVar->setOffset(field->getOffset());
 
    if ( field->getType()->isUserType() )
      declareUserTypeFieldVars(fieldVar);
  }

  program->leaveScope();
}



