/********************************************************************

    project C-Control II Compiler

    parser.cpp

********************************************************************/


#include "parser.h"
#include "error.h"
#include "text.h"



//-----------------------------------------------------------------------
  long Parser::run ( Program* _program, Module* _module, 
                     std::ostream* _errLog, std::string  _errLineHeader )
//-----------------------------------------------------------------------
{
  program             = _program;
  module              = _module;
  input               = module->getTokens();
  errLog              = _errLog;
  errLineHeader       = _errLineHeader;
  tokenIterator       = input->gotoToken(0);
  lineNo              = 1;
  currentlyParsedProc = 0;
  errCount            = 0;
  isParsingConstTerm  = false;

  lineinfos = module->getLineInfos();
  lineinfos->clear();
  lineinfos->push_back(Module::LINE_EMPTY);
  
  try
  {
    parseModule(); 
  }	
  catch (...)
  {
  }

  return errCount;
}


//----------------------------
  void Parser::savePosition ()
//----------------------------
{
  prev_lineNo = lineNo;
  prev_tokenIterator = tokenIterator;
}


//-------------------------------
  void Parser::restorePosition ()
//-------------------------------
{
  lineNo = prev_lineNo;
  tokenIterator = input->gotoToken(prev_tokenIterator);
}


//-----------------------------
  void Parser::markErrorLine ()
//-----------------------------
{
  (*lineinfos)[lineNo-1] |= Module::LINE_ERROR;
}


//----------------------------------
  void Parser::markExecutableLine ()
//----------------------------------
{
  (*lineinfos)[lineNo-1] |= Module::LINE_EXECUTABLE;

  long instructionID = program->logInstruction(module->getHandle(), lineNo);

  if ( instructionID < 0 )
  {
    complain(Text::get(Text::PROGRAM_TOO_BIG));
    throw Error(Error::END_OF_FILE);
  }

  putCodeWithID(IntermediateCode::ENTER_INSTRUCTION, (unsigned)instructionID);
}



//----------------------------------------------------
  Token::ID Parser::readNextToken ( Token::ID symbol )
//----------------------------------------------------
{
  Token::ID id;

  for (;;) 
  {
    tokenIterator = input->gotoNextToken();
    skipLineFeeds();
    id = input->getToken().id;

    if ( id == Token::INVALID )
    {
  	  std::string errMsg;
      errMsg = Text::get(Text::INVALID_SYMBOL);
      errMsg += " ";
      errMsg += Text::get(symbol);
      complain(errMsg);
    }
    else if ( id == Token::UNEXPECTED_END_OF_LINE )
    {
  	  std::string errMsg;
      errMsg = Text::get(Text::UNEXPECTED);
      errMsg += " ";
      errMsg += Text::get(Text::LINE_FEED);
      complain(errMsg);
    }
    else
      break;
  }

  if (symbol)
    expect(symbol);

  return id;
}


//----------------------------
  void Parser::putBackToken ()
//----------------------------
{
  tokenIterator = input->putBackToken();
}


//-----------------------------
  void Parser::skipLineFeeds ()
//-----------------------------
{
  for (;;)
  {
    if ( input->getToken().id == Token::LINE_FEED )
    {
      ++lineNo;
      tokenIterator = input->gotoNextToken();
      lineinfos->push_back(Module::LINE_EMPTY);
    }
    else
      break;
  }
}


//--------------------------------------------
  void Parser::synchronize ( Token::ID synch )
//--------------------------------------------
{
  do 
  {
    if ( input->getToken().id == Token::END_OF_FILE )
    {
      std::string errMsg;
	    errMsg = Text::get(Text::UNEXPECTED);
      errMsg += " ";
      errMsg += Text::get(Text::END_OF_FILE);
      complain(errMsg);  
      throw Error(Error::END_OF_FILE);
    }
    
  }	while (	readNextToken() != synch );

  putBackToken();
}


//----------------------------------------------------------------------
  void Parser::synchronize ( const Token::ID* pSynch, int numSynch )
//----------------------------------------------------------------------
{
  bool foundSynchronizer = false;
 
  do
  {
    if ( input->getToken().id == Token::END_OF_FILE )
    {
      std::string errMsg;
	    errMsg = Text::get(Text::UNEXPECTED);
      errMsg += " ";
      errMsg += Text::get(Text::END_OF_FILE);
      complain(errMsg);  
      throw Error(Error::END_OF_FILE);
    }
    
    readNextToken();

    for ( int i = 0; i < numSynch; i++ )
    {
      if ( input->getToken().id == pSynch[i] )
      {
        foundSynchronizer = true;
        break;
      }
    }
  }	
  while ( !foundSynchronizer );

  putBackToken();
}


// --- error checking and messaging ---------------------------------

//----------------------------------------------
  void Parser::unexpect ( Token::ID symbol )
//----------------------------------------------
{
  std::string errMsg;
  errMsg = Text::get(Text::UNEXPECTED);
  errMsg += " ";
  errMsg += Text::get(symbol);
  complain(errMsg);
  throw Error();
}



//----------------------------------------
  void Parser::expect ( Token::ID symbol )
//----------------------------------------
{
  if ( input->getToken().id != symbol )
  {
    std::string errMsg;
    errMsg  = Text::get(symbol);
    errMsg += " ";
    errMsg += Text::get(Text::EXPECTED);
    complain(errMsg);  
    throw Error();
  }
}


//-----------------------------------------------------------------------
  void Parser::expect ( Identifier* identifier, const std::string& name )
//-----------------------------------------------------------------------
{
  if ( !identifier )
  {
    std::string errMsg;
    errMsg = Text::get(Text::UNKNOWN);
    errMsg += " ";
    errMsg += Text::get(Text::IDENTIFIER);
    errMsg += " \"";
    errMsg += name;
    errMsg += "\"";
    complain(errMsg);
    throw Error();
  }
}


//------------------------------------------------------------------
  void Parser::expect ( Identifier* identifier, IDENTCLASS idclass )
//------------------------------------------------------------------
{
  Identified* theIdentifiedThing;
  bool isGood = false;

  if ( identifier )
  {
    theIdentifiedThing = dynamic_cast<Identified*>(identifier->getValue().get());

    if ( theIdentifiedThing->what() == idclass )
      isGood = true;

    if ( DATA_IDENT == idclass )
      if ( theIdentifiedThing->what() >  DATA_IDENT )
        isGood = true;
  }

  if ( !isGood )
  {
    std::string errMsg;

    switch ( idclass )
    {
       case THREAD_IDENT   : errMsg = Text::get(Text::THREAD_IDENTIFIER); break;
       case FUNCTION_IDENT : errMsg = Text::get(Text::FUNCTION_IDENTIFIER); break;
       case TYPE_IDENT     : errMsg = Text::get(Text::TYPE_IDENTIFIER); break;
       case VAR_IDENT      : errMsg = Text::get(Text::VARIABLE_IDENTIFIER); break;

       default : errMsg = Text::get(Text::CONSTANT_IDENTIFIER); break;
    }

    errMsg += " ";
    errMsg += Text::get(Text::EXPECTED);
    complain(errMsg);  
    throw Error();
  }
}


//---------------------------------------------------
  void Parser::checkType ( Type* type1, Type* type2 )
//---------------------------------------------------
{
  if ( type1 != type2 )
  {
    std::string errMsg = Text::get(Text::TYPE_MISMATCH);
    complain(errMsg);  
    throw Error(Error::TYPE_MISMATCH);
  }  
}

/*
//---------------------------------------------------
  void Parser::checkType ( Data* data1, Data* data2 )
//---------------------------------------------------
{
  if ( 
        (data1->getType() != data2->getType()) || 
        (data1->isArray() != data2->isArray()) 
     )
  {
    std::string errMsg = Text::get(Text::TYPE_MISMATCH);
    complain(errMsg);  
    throw Error(Error::TYPE_MISMATCH);
  }  
}
*/

//---------------------------------------------
  void Parser::expectNumericType ( Type* type )
//---------------------------------------------
{
  if ( 
       ( type != &StandardType::BYTE ) &&
       ( type != &StandardType::INT ) &&
       ( type != &StandardType::LONG ) &&
       ( type != &StandardType::FLOAT ) 
     )
  {
    std::string errMsg = Text::get(Text::TYPE_MISMATCH);
    complain(errMsg);  
    throw Error(Error::TYPE_MISMATCH);
  }
}


//-----------------------------------------------------------
  Function* Parser::expectInsideFunction ( Token::ID symbol )
//-----------------------------------------------------------
{
  bool isInsideFunction = false;

  if ( isLocal() ) 
    isInsideFunction = (currentlyParsedProc->what() == FUNCTION_IDENT);
  
  if ( !isInsideFunction )
  {
    std::string errMsg;
    errMsg = Text::get(Text::UNEXPECTED);
    errMsg += " ";
    errMsg += Text::get(symbol);
    complain(errMsg);
    throw Error();
  }

  return dynamic_cast<Function*>(currentlyParsedProc);
}


//-------------------------------------------------------
  void Parser::complain ( const std::string& whatsWrong )
//-------------------------------------------------------
{
  errCount++;

  if ( errLog )
    (*errLog ) << errLineHeader << " (" << lineNo << ") : " << whatsWrong << std::endl; 

  markErrorLine();
}


// -----------------


//--------------------------------------------------------------------------
  Identifier* Parser::declare ( const std::string& name, Identified* object, 
                                IDENTCLASS idclass, bool hide )
//--------------------------------------------------------------------------
{
  Identifier* identifier = program->declare(declarationLineNo, name, object, idclass, hide);

  if ( ! identifier  )
  {
    std::string errMsg;
    errMsg = Text::get(Text::MULTIPLE_DECLARATION);
    errMsg += " \"";
    errMsg += name;
    errMsg += "\"";   
    complain(errMsg);  
    throw Error();
  }

  return identifier;
}


//---------------------------
  void Parser::parseModule ()
//---------------------------
{
  static const Token::ID synch[] = { Token::THREAD, Token::FUNCTION, Token::TYPE, 
                                     Token::CONSTANT, Token::BYTE, Token::INT, Token::LONG, 
                                     Token::FLOAT }; 

  long prevErrCount = errCount;

  for ( skipLineFeeds(); input->getToken().id != Token::END_OF_FILE; readNextToken() )
  {
    if ( input->getToken().id == Token::SEMICOLON )
      continue;

    Identifier* identifier;
   
    try
    {   
      identifier = program->getScope();
      currentlyParsedProc = 0;

      switch ( input->getToken().id  )
      {
        case Token::THREAD   : parseThread(); break;
        case Token::INLINE   : readNextToken(Token::FUNCTION); parseFunction (true); break;
        case Token::FUNCTION : parseFunction (); break;
        case Token::TYPE     : parseUserTypeDef (); break;
        case Token::CONSTANT : parseConstDef (); break;

        default: parseVarDef (); break; 
      }

      if ( input->getToken().id != Token::RIGHT_BRACE )
        readNextToken(Token::SEMICOLON);
    }
    catch ( Error& err )
    {
      if ( err.id == Error::END_OF_FILE )
        throw err;

      synchronize(synch, sizeof(synch)/sizeof(synch[0]));
      program->restoreScope(identifier);
    }
  }

  if ( prevErrCount == errCount )
    (*errLog ) << errLineHeader << " : OK" << std::endl; 
}


//--------------------------------------
  Identifier* Parser::parseIdentifier ()
//--------------------------------------
{
  Identifier* identifier;

  expect(Token::IDENTIFIER);
  identifier = program->getScope()->seekChild(input->getString());
  expect(identifier, input->getString());

  if ( (dynamic_cast<Identified*>(identifier->getValue().get()))->what() == MODULE_IDENT )
  {
    readNextToken(Token::DOT);
    readNextToken(Token::IDENTIFIER);
    identifier = identifier->getChild(input->getString());
    expect(identifier, input->getString());
  }

  return identifier;
}


//----------------------------------------------------------------------------------------------------------
  Data* Parser::parseDataAccess ( Data* data, bool& isIndirect, bool isConstAllowed, bool shallLoadPointer )
//----------------------------------------------------------------------------------------------------------
{
  if ( !isConstAllowed )
    expect(data->getIdentifier(), VAR_IDENT);

  if ( data->isArray() )
  {  
    if ( shallLoadPointer ) 
    {
      putCodeWithObject(IntermediateCode::LOAD_PTR, data );
      isIndirect = true;
    }

    if ( readNextToken() == Token::LEFT_BRACKET )
    {
      parseArrayElementAccess(data);

      if ( data->isUserType() )
      {
        if ( readNextToken() == Token::DOT )
        {
          data = parseUserTypeFieldAccess(data);
          data = parseDataAccess (data, isIndirect, false, false );
        }
        else
        {
          putBackToken();
        }
      }
    }
    else
    {
      putBackToken();
    }
  }
  else if ( data->isUserType() )
  {
    if ( readNextToken() == Token::DOT )
    {
      if ( shallLoadPointer ) 
      {
        putCodeWithObject(IntermediateCode::LOAD_PTR, data );
        isIndirect = true;
      }

      data = parseUserTypeFieldAccess(data);
      data = parseDataAccess ( data, isIndirect, false, false );
    }
    else
    {
      putBackToken();
    }
  }
  else if ( data->getType()->isStringType() )
  {
    if ( shallLoadPointer ) 
    {
      putCodeWithObject(IntermediateCode::LOAD_PTR, data );
      isIndirect = true;
    }
  }

  return data;
}


//---------------------------------------------------
  void Parser::parseArrayElementAccess ( Data* data )
//---------------------------------------------------
{
  Identifier* identifier = data->getIdentifier();

  readNextToken();
  parseExpression();
  readNextToken(Token::RIGHT_BRACKET);  

  if ( data->getMemoryClass() == Data::CONSTANT )
  {
    if ( data->getType() == &StandardType::STRING )
      putImmLoadValueCode(1);
    else
      putImmLoadValueCode(data->getType()->getSize() / 2); // xxx backend detail, better remove from here!
  }
  else
    putImmLoadValueCode(data->getType()->getSize());

  putCode(IntermediateCode::MUL);
  putCode(IntermediateCode::ADD);
}


//-----------------------------------------------------
  Data* Parser::parseUserTypeFieldAccess ( Data* data )
//-----------------------------------------------------
{
  readNextToken(Token::IDENTIFIER);
  Identifier* identifier = data->getIdentifier()->getChild(input->getString());
  expect(identifier, input->getString());
  data = dynamic_cast<Data*>(identifier->getValue().get());
  putImmLoadValueCode(data->getOffset());
  putCode(IntermediateCode::ADD);

  return data;
}


//--------------------------------
  double Parser::parseConstTerm ()
//--------------------------------
{
  double value;
  AutomaticFlag automaticflag(isParsingConstTerm);

  IntermediateCodeSequence constseq;
  enterCodeSeq(&constseq);

  readNextToken();
  parseExpression();

  try
  {
    value = constseq.getConstantResult();
  }
  catch (...)
  {
    std::string errMsg;
    errMsg = Text::get(Text::CONSTANT_EXPRESSION);
    errMsg += " ";
    errMsg += Text::get(Text::EXPECTED);

    complain(errMsg);
    leaveCodeSeq();
    throw Error();
  }

  leaveCodeSeq();
  return value; 
}


//--------------------------------------------------------------
  void Parser::restoreCodeSeq ( IntermediateCodeSequence* _seq )
//--------------------------------------------------------------
{
  while ( seq.top() != _seq )
    seq.pop();
}


//-------------------------------------------------------------------------------------------------
  void Parser::chainCodeSeq ( IntermediateCodeSequence* _seq, bool shallCopyInstructionChangeInfo )
//-------------------------------------------------------------------------------------------------
{ 
  _seq->finish();
  seq.top()->chainCodeSeq(_seq, shallCopyInstructionChangeInfo);
}
