/********************************************************************

    project C-Control II Compiler

    parser.h

********************************************************************/

#ifndef parser_h
#define parser_h


#include "classlib/classlib.h"
#include "token.h"

#include "token.h"
#include "module.h"
#include "procedure.h"
#include "program.h"


//------------
  class Parser
//------------
{
  protected:

    // --- data interface ---

    Program*           program;
    Module*            module;
    ScanResult*        input;
    Module::LineInfos* lineinfos;
    Procedure*         currentlyParsedProc;
  
  	// --- sequential reading of tokens  ---

    unsigned long tokenIterator;
    unsigned long prev_tokenIterator;

    void savePosition ();
    void restorePosition ();

    void markErrorLine ();
    void markExecutableLine ();

/*    Token* token() const;
    StringToken* stringtoken() const;
    ValueToken* valuetoken() const;
    FloatValueToken* floatvaluetoken() const; */

    Token::ID readNextToken ( Token::ID symbol = Token::NONE );

    void putBackToken ();
    void skipLineFeeds ();
    void synchronize ( Token::ID synch = Token::NONE );
    void synchronize ( const Token::ID* pSynch, int numSynch );

    // --- error checking and messaging ---

    std::ostream* errLog;
    std::string   errLineHeader;
    int lineNo;
    int prev_lineNo;
    int declarationLineNo;
    long errCount;

    void unexpect ( Token::ID symbol );
    void expect ( Token::ID symbol );
    void expect ( Identifier* identifier, const std::string& name );
    void expect ( Identifier* identifier, IDENTCLASS idclass = DATA_IDENT );

    void checkType ( Type* type1, Type* type2 );    
//    void checkType ( Data* data1, Data* data2 );    
    void expectNumericType ( Type* type );    
    Function* expectInsideFunction ( Token::ID symbol );

    void complain ( const std::string& whatsWrong );

    // --- recursive top-down parsing ---

    bool isLocal ();

    Identifier* declare ( const std::string& name, Identified* object, IDENTCLASS idclass, bool hide = false );
    void declareUserTypeFieldVars ( Variable* variable );

    void parseModule ();
    void parseThread ();
    void parseFunction ( bool inlineFunction = false );
    void parseUserTypeDef ();
    void parseConstDef ();
    void parseVarDef ();

    void parseInlineBlock ();
    void parseInstructionBlock ();
    void parseInstruction ();
    void parseSimpleInstruction ();
    void parseActualParameter ( Function* pFunction );

    void parseExpression ();
    void parseTerm ();
    void parseOr ();
    void parseAnd ();
    void parseCompare ();
    void parseShift ();
    void parseSum ();
    void parseProduct ();
    void parseOperand ();

    void parseStringConcatenation ( Variable* stringVar );
    void parseStringOperand ( Variable* stringVar );

    void parseReturn ();
    void parseBreak ();
    void parseContinue ();
    
    void parseInline ();
    void parseRun ();
    void parseHalt ();
    void parseResume ();
    void parseReset ();
    void parseSleep ();
    void parseQuit ();
    void parseCapture ();
    void parseRelease ();
        
  	void parseWait ();
    void parseLoop ();
    void parseWhile ();
    void parseDo ();
    void parseIf ();
    void parseOn ();
    void parseFor ();

    Identifier* parseIdentifier ();
    Data* parseDataAccess ( Data* data, bool& isIndirect, bool isConstAllowed = true, bool shallLoadPointer = true );
    Data* parseUserTypeFieldAccess ( Data* data );
    void  parseArrayElementAccess ( Data* data );

    Type* parseType ();

    double parseConstTerm ();
    bool   isParsingConstTerm;
    int    iTempString;

    unsigned parseNameWithBrackets ( std::string& name );

    // --- intermediate code generation ---

    std::stack<IntermediateCodeSequence*> seq;

    void enterCodeSeq ( IntermediateCodeSequence* _seq );
    void leaveCodeSeq ();
    void restoreCodeSeq ( IntermediateCodeSequence* _seq );
    void chainCodeSeq ( IntermediateCodeSequence* _seq, bool shallCopyInstructionChangeInfo = true );

    void putCode ( IntermediateCode::OPERATION opcode ); 
    void putCodeWithID ( IntermediateCode::OPERATION opcode, unsigned id );
    void putImmLoadValueCode ( long value );
    void putImmLoadFloatValueCode ( double value );
    void putCodeWithObject ( IntermediateCode::OPERATION opcode, Virtual* object );
    void putCodeWithString ( IntermediateCode::OPERATION opcode, const std::string& string );

    std::stack<unsigned> breakMarks;
    std::stack<unsigned> continueMarks;

  public:

    long run ( Program* _pProgram, Module* _module, std::ostream* _errLog, std::string  _errLineHeader );
};


/*inline Token* Parser::token() const
{ return (*tokenIterator).get(); }

inline StringToken* Parser::stringtoken() const
{ return (StringToken*)token(); }

inline ValueToken* Parser::valuetoken() const
{ return (ValueToken*)token(); }

inline FloatValueToken* Parser::floatvaluetoken() const
{ return (FloatValueToken*)token(); }
*/

inline bool Parser::isLocal ()
{ return currentlyParsedProc != 0; }

inline void Parser::enterCodeSeq ( IntermediateCodeSequence* _seq )
{ seq.push(_seq); }

inline void Parser::leaveCodeSeq ()
{ seq.pop(); }

inline void Parser::putCode ( IntermediateCode::OPERATION opcode )
{ seq.top()->putCode(opcode); }

inline void Parser::putCodeWithID ( IntermediateCode::OPERATION opcode, unsigned id )
{ seq.top()->putCodeWithID(opcode, id); }

inline void Parser::putImmLoadValueCode ( long value )
{ seq.top()->putImmLoadValueCode(value); }

inline void Parser::putImmLoadFloatValueCode ( double value )
{ seq.top()->putImmLoadFloatValueCode(value); }

inline void Parser::putCodeWithObject ( IntermediateCode::OPERATION opcode, Virtual* object )
{ seq.top()->putCodeWithObject(opcode, object); }

inline void Parser::putCodeWithString ( IntermediateCode::OPERATION opcode, const std::string& string )
{ seq.top()->putCodeWithString(opcode, string); }

#endif
