/********************************************************************

    project C-Control II Compiler

    parsfunc.cpp

********************************************************************/

#include "parser.h"
#include "error.h"
#include "text.h"



//---------------------------
  void Parser::parseThread ()
//---------------------------
{
  declarationLineNo = lineNo;
  readNextToken(Token::IDENTIFIER); 
  std::string name = input->getString();

  Thread* thread = new Thread( name == "main" ? Thread::NORMAL_PRIORITY : Thread::IDLE_PRIORITY);
  Identifier* identifier = declare(name, thread, THREAD_IDENT, true);
  program->enterScope(identifier);
  module->changeContext(lineNo, identifier);
  enterCodeSeq(thread->getImCodeSeq());
  currentlyParsedProc = thread;

  unsigned mark = seq.top()->mark();
  readNextToken();
  parseInstructionBlock ();
  putCodeWithID(IntermediateCode::GOTO, mark);

  seq.top()->requireMemory(96); // reserve memory for operands stack
  seq.top()->requireMemory(thread->getLocalMemSize());

  program->leaveScope();
  leaveCodeSeq();
}



//--------------------------------------------------
  void Parser::parseFunction ( bool inlineFunction )
//--------------------------------------------------
{
  int numFormalParams = 0;

  declarationLineNo = lineNo;
  readNextToken(Token::IDENTIFIER); 
  std::string name = input->getString();

  Function* function = new Function(inlineFunction);
  Identifier* identifier = declare(name, function, FUNCTION_IDENT);

  if ( !inlineFunction )
  {
    Variable* capvar = new Variable(&StandardType::BYTE, Data::STATIC); 
    declare(name + "_@cap", capvar, VAR_IDENT);
    capvar->declareSecret();
    function->setCapVar(capvar);
  }

  program->enterScope(identifier);
  module->changeContext(lineNo, identifier);
  enterCodeSeq(function->getImCodeSeq());
  currentlyParsedProc = function;

  readNextToken(Token::LEFT_PARENTHESIS);
  
  if ( readNextToken() != Token::RIGHT_PARENTHESIS )
  {
    putBackToken();

    do 
    {
      bool isArrayParam = false;

      readNextToken();
      Type* type = parseType();
 
      declarationLineNo = lineNo;
      readNextToken(Token::IDENTIFIER);
      std::string paramName = input->getString();
 
      if ( readNextToken() == Token::LEFT_BRACKET )
      {   
        isArrayParam = true;
        readNextToken(Token::RIGHT_BRACKET);
      }
      else
      { 
        if ( type == &StandardType::BYTE )  // bytes as int if no array
          type = &StandardType::INT;
        putBackToken();
      }

      if ( type->isNumericType() && !isArrayParam )
      {
        Variable* variable = new Variable(type, Data::LOCAL); 
        declare(paramName, variable, VAR_IDENT);
        function->addParameter(variable);
      }
      else
      {
        Reference* reference = new Reference(type, isArrayParam); 
        declare(paramName + "_@ref", reference, VAR_IDENT);
        function->addParameter(reference);

        Variable* variable = new Variable(type, Data::LOCAL, isArrayParam, 0, reference); 
        declare(paramName, variable, VAR_IDENT);
        function->addLocalVariable(variable);
        if ( type->isUserType() )
          declareUserTypeFieldVars(variable);
      }
      numFormalParams++;

    } while ( readNextToken() == Token::COMMA ); 

    expect(Token::RIGHT_PARENTHESIS); 
  }
 
  if ( readNextToken() == Token::RETURNS )
  {
    readNextToken();
    Type* type = parseType();

    if ( type == &StandardType::BYTE )  // !!!
      type = &StandardType::INT;

    function->setType(type);
    readNextToken();
  }

  function->locateParameters();

  if ( inlineFunction )
  {
    parseInlineBlock();
  }
  else
  {
    parseInstructionBlock ();

    if ( function->getType() )
      putImmLoadValueCode(0);
    
    putCodeWithObject(IntermediateCode::RETURN, function);
  }

  seq.top()->requireMemory(96); // reserve memory for operands stack
  seq.top()->requireMemory(function->getLocalMemSize());
  seq.top()->requireMemory(function->getParamMemSize());

  leaveCodeSeq();
  program->leaveScope();
  module->changeContext(lineNo, program->getScope());
}


//--------------------------------------------------------
  void Parser::parseActualParameter ( Function* function )
//--------------------------------------------------------
{
  readNextToken(Token::LEFT_PARENTHESIS);

  int numParams = function->getNumParams();

  for ( int i = 0; i < numParams; i++ )
  {
    if ( i )
      readNextToken(Token::COMMA);

    Data* parameter = function->getParameter(i);
    Type* type = parameter->getType();

    readNextToken();

    if ( type->isNumericType() )
    {
      if ( parameter->isArray() )
      {
        bool dummy;

        expect(Token::IDENTIFIER);
        Identifier* identifier = parseIdentifier();
        Identified* identified = dynamic_cast<Identified*>(identifier->getValue().get());
        expect(identifier);
        Data* actualparam  = dynamic_cast<Data*>(identified);

        actualparam->use();
        putCodeWithObject(IntermediateCode::LOAD_PTR, actualparam);
        actualparam = dynamic_cast<Variable*>(parseDataAccess(actualparam, dummy, false, false));

        bool stringAsByteArray = (type == &StandardType::BYTE) &&
                                (actualparam->getType() == &StandardType::STRING);
        
        if ( !stringAsByteArray )
          checkType(parameter->getType(), actualparam->getType());
      }
      else
      {
        parseExpression();
        putCodeWithID(IntermediateCode::CAST, (unsigned) (dynamic_cast<NumericType*>(type))->getForm());
      }
    }
    else if ( type->isUserType() )
    {
      bool dummy;

      expect(Token::IDENTIFIER);
      Identifier* identifier = parseIdentifier();
      Identified* identified = dynamic_cast<Identified*>(identifier->getValue().get());
      expect(identifier);
      Data* actualparam  = dynamic_cast<Data*>(identified);

      actualparam->use();
      putCodeWithObject(IntermediateCode::LOAD_PTR, actualparam);
      actualparam = dynamic_cast<Variable*>(parseDataAccess(dynamic_cast<Data*>(identified), dummy, false, false));

      checkType(parameter->getType(), actualparam->getType());
    }

    else if ( type->isStringType() )
    {
      if ( parameter->isArray() )
      {
        expect(Token::IDENTIFIER);
        Identifier* identifier = parseIdentifier();
        expect(identifier, VAR_IDENT);
        Data* actualparam = dynamic_cast<Data*>(identifier->getValue().get());
        actualparam->use();
        putCodeWithObject(IntermediateCode::LOAD_PTR, actualparam);
        checkType(parameter->getType(), actualparam->getType());
      }
      else
      {
        Variable* tempString;

        if ( input->getToken().id == Token::LITERAL_STRING )
        {
          tempString = currentlyParsedProc->getTempString(iTempString++);
          putCodeWithObject(IntermediateCode::LOAD_PTR, tempString);
          putCode(IntermediateCode::STR_CLEAR);
          putCodeWithObject(IntermediateCode::LOAD_PTR, tempString);
          putCodeWithString(IntermediateCode::STR_PUTISTR, input->getString());
          putCodeWithObject(IntermediateCode::LOAD_PTR, tempString);
        }
        else
        {
          expect(Token::IDENTIFIER);
          Identifier* identifier = parseIdentifier();
          expect(identifier);
          Data* actualparam = dynamic_cast<Data*>(identifier->getValue().get());
          checkType(parameter->getType(), actualparam->getType());
          actualparam->use();

          if ( actualparam->getMemoryClass() == Data::CONSTANT )
          {
            tempString = currentlyParsedProc->getTempString(iTempString++);
            putCodeWithObject(IntermediateCode::LOAD_PTR, tempString);
            putCode(IntermediateCode::STR_CLEAR);
            putCodeWithObject(IntermediateCode::LOAD_PTR, tempString);
          }

          putCodeWithObject(IntermediateCode::LOAD_PTR, actualparam);

          if ( actualparam->isArray() )
          {
            if ( readNextToken() == Token::LEFT_BRACKET ) 
              parseArrayElementAccess(actualparam);
            else
              putBackToken();
          }

          if ( actualparam->getMemoryClass() == Data::CONSTANT )
          {
            putCodeWithObject(IntermediateCode::STR_PUTCSTR, actualparam->isArray()? actualparam : (Data*)0);
            putCodeWithObject(IntermediateCode::LOAD_PTR, tempString);
          }
        }
      }
    }
  }

  readNextToken(Token::RIGHT_PARENTHESIS);
}


//---------------------------
  void Parser::parseReturn ()
//---------------------------
{
  Function* function = expectInsideFunction(Token::RETURN);
  
  if ( function->getType() )
  {
    readNextToken();
  	parseExpression();
  }

  putCodeWithObject(IntermediateCode::RETURN, function);
}

