/********************************************************************

    project C-Control II Compiler

    parsinst.cpp

********************************************************************/


#include "parser.h"
#include "error.h"
#include "text.h"


//--------------------------------
  void Parser::parseInlineBlock ()
//--------------------------------
{
  static const Token::ID synch[] = { Token::SEMICOLON, Token::RIGHT_BRACE }; 

  Identifier* oldScope;
  IntermediateCodeSequence* oldSeq;

  expect(Token::LEFT_BRACE);

  while ( readNextToken() != Token::RIGHT_BRACE )
  {
    if ( input->getToken().id == Token::SEMICOLON )
      continue;

    try
    {
      oldScope = program->getScope();
      oldSeq = seq.top();

      expect(Token::INLINE);
      parseInline(); 

      if ( input->getToken().id != Token::RIGHT_BRACE )
        readNextToken(Token::SEMICOLON);
    }
    catch ( Error& err )
    {
      if ( err.id == Error::END_OF_FILE )
        throw err;

      synchronize(synch, sizeof(synch)/sizeof(synch[0]));
      program->restoreScope(oldScope);
      restoreCodeSeq(oldSeq);
    }
  }
}


//-------------------------------------
  void Parser::parseInstructionBlock ()
//-------------------------------------
{
  static const Token::ID synch[] = { Token::SEMICOLON, Token::RIGHT_BRACE }; 

  Identifier* oldScope;
  IntermediateCodeSequence* oldSeq;

  expect(Token::LEFT_BRACE);

  while ( readNextToken() != Token::RIGHT_BRACE )
  {
    if ( input->getToken().id == Token::SEMICOLON )
      continue;

    try
    {
      oldScope = program->getScope();
      oldSeq = seq.top();

      parseInstruction(); 

      if ( input->getToken().id != Token::RIGHT_BRACE )
        readNextToken(Token::SEMICOLON);
    }
    catch ( Error& err )
    {
      if ( err.id == Error::END_OF_FILE )
        throw err;

      synchronize(synch, sizeof(synch)/sizeof(synch[0]));
      program->restoreScope(oldScope);
      restoreCodeSeq(oldSeq);
    }
  }
}


//--------------------------------
  void Parser::parseInstruction ()
//--------------------------------
{
  std::string string;

  iTempString = 0;

  switch( input->getToken().id )
  {
    case Token::SEMICOLON:   putBackToken(); break;

    case Token::LEFT_BRACE:

      parseInstructionBlock();
      break;

    case Token::RETURN   : markExecutableLine(); parseReturn(); break;
    case Token::BREAK    : markExecutableLine(); parseBreak(); break;
    case Token::CONTINUE : markExecutableLine(); parseContinue(); break;
    case Token::IF       : markExecutableLine(); parseIf(); break;
    case Token::FOR      : markExecutableLine(); parseFor(); break;
    case Token::WHILE    : markExecutableLine(); parseWhile(); break;
    case Token::DO       : parseDo(); break;
    case Token::LOOP     : parseLoop(); break;
    case Token::WAIT     : markExecutableLine(); parseWait(); break;

    case Token::INLINE   : markExecutableLine(); parseInline(); break;

    case Token::RUN      : markExecutableLine(); parseRun(); break;
    case Token::HALT     : markExecutableLine(); parseHalt(); break;
    case Token::RESUME   : markExecutableLine(); parseResume(); break;
    case Token::RESET    : markExecutableLine(); parseReset(); break;
    case Token::SLEEP    : markExecutableLine(); parseSleep(); break;
  
    case Token::CAPTURE  : markExecutableLine(); parseCapture(); break;
    case Token::RELEASE  : markExecutableLine(); parseRelease(); break;
    case Token::YIELD    : markExecutableLine(); putCode(IntermediateCode::YIELD); break;
    case Token::QUIT     : markExecutableLine(); parseQuit(); break;

    default: parseSimpleInstruction();
  }
}


//--------------------------------------
  void Parser::parseSimpleInstruction ()
//--------------------------------------
{
  if ( input->getToken().id == Token::IDENTIFIER )
  {
    savePosition();

    Identifier* identifier = parseIdentifier();
    Identified* identified = dynamic_cast<Identified*>(identifier->getValue().get());

    bool isIndirect = false;
    Variable* variable;
    IntermediateCodeSequence accessSeq;
    Function* function;

    switch ( identified->what() )
    {
      case VAR_IDENT :

        identified->use();
        markExecutableLine(); 
        enterCodeSeq(&accessSeq);
        variable = dynamic_cast<Variable*>(parseDataAccess(dynamic_cast<Data*>(identified), isIndirect, false));
        leaveCodeSeq();

        readNextToken(Token::ASSIGN);
        readNextToken();

        if ( variable->getType()->isStringType() )
        {
           Variable* tempString;

           tempString = currentlyParsedProc->getTempString(iTempString++);
           putCodeWithObject(IntermediateCode::LOAD_PTR, tempString);
           putCode(IntermediateCode::STR_CLEAR);

           parseStringConcatenation(tempString);
           chainCodeSeq(&accessSeq);
           putCode(IntermediateCode::STR_CLEAR);
           chainCodeSeq(&accessSeq);
           putCodeWithObject(IntermediateCode::LOAD_PTR, tempString);
           putCode(IntermediateCode::STR_PUTSTR);
        }
        else
        {
          expectNumericType(variable->getType());
          parseExpression();
          putCodeWithID(IntermediateCode::CAST, dynamic_cast<NumericType*>(variable->getType())->getForm());
          chainCodeSeq(&accessSeq);
          putCodeWithObject(isIndirect? IntermediateCode::R_STORE : IntermediateCode::STORE, variable);
        }
          
        break;

      case FUNCTION_IDENT : 
       
        markExecutableLine(); 
        function = dynamic_cast<Function*>(identified);
        function->use();
        parseActualParameter(function);
 
        if ( function->isInline() )
          chainCodeSeq (function->getImCodeSeq(), false);
        else
          putCodeWithObject(IntermediateCode::CALL, function);

        if ( function->getType() )
          putCodeWithObject(IntermediateCode::WIPESTACK, function);

        seq.top()->requireMemoryForCall(function->getImCodeSeq()->getMemoryRequirements());

        break;

      default :
  
        restorePosition();
        parseVarDef();
    }
  }
  else
    parseVarDef();
}




//-------------------------------
  void Parser::parseExpression ()
//-------------------------------
{
  seq.top()->beginConstantFolding();
  parseTerm();
}



