/********************************************************************

    project C-Control II Compiler

    parser.cpp

********************************************************************/


#include "parser.h"
#include "error.h"
#include "text.h"



//--------------------------
  void Parser::parseBreak ()
//--------------------------
{
  if ( breakMarks.empty() )
    unexpect(Token::BREAK);
  else
    putCodeWithID(IntermediateCode::GOTO, breakMarks.top());
}


//-----------------------------
  void Parser::parseContinue ()
//-----------------------------
{
  if ( continueMarks.empty() )
    unexpect(Token::CONTINUE);
  else
    putCodeWithID(IntermediateCode::GOTO, continueMarks.top());
}


//---------------------------
  void Parser::parseInline ()
//---------------------------
{
  putCodeWithID(IntermediateCode::INLINE, (unsigned)parseConstTerm());
}


//------------------------
  void Parser::parseRun ()
//------------------------
{ 
  if ( readNextToken() == Token::IDENTIFIER )
  {
    savePosition();
    Identifier* identifier = parseIdentifier();
    Identified* identified = dynamic_cast<Identified*>(identifier->getValue().get());

    if ( identified->what() == THREAD_IDENT )
    {
      putImmLoadValueCode(Thread::NORMAL_PRIORITY);
      putCodeWithObject(IntermediateCode::RUN, identified);
      return;
    }
    restorePosition();    
  }

  parseExpression();
  putCodeWithID(IntermediateCode::CAST, NumericType::INT);
  putCodeWithObject(IntermediateCode::RUN, (Identified*) 0); 
}


//-------------------------
  void Parser::parseHalt ()
//-------------------------
{ 
  if ( readNextToken() == Token::IDENTIFIER )
  {
    Identifier* identifier = parseIdentifier();
    expect(identifier, THREAD_IDENT);
    putCodeWithObject(IntermediateCode::HALT, dynamic_cast<Identified*>(identifier->getValue().get()));
  }
  else
  {
    putBackToken();
    putCodeWithObject(IntermediateCode::HALT, (Virtual*) 0); 
  }
}


//---------------------------
  void Parser::parseResume ()
//---------------------------
{ 
  if ( readNextToken() == Token::IDENTIFIER )
  {
    Identifier* identifier = parseIdentifier();
    expect(identifier, THREAD_IDENT);
    putCodeWithObject(IntermediateCode::RESUME, dynamic_cast<Identified*>(identifier->getValue().get()));
  }
  else
  {
    putBackToken();
    putCodeWithObject(IntermediateCode::RESUME, (Virtual*) 0); 
  }
}


//--------------------------
  void Parser::parseReset ()
//--------------------------
{ 
  if ( readNextToken() == Token::IDENTIFIER )
  {
    Identifier* identifier = parseIdentifier();
    expect(identifier, THREAD_IDENT);
    putCodeWithObject(IntermediateCode::RESET, dynamic_cast<Identified*>(identifier->getValue().get()));
  }
  else
  {
    putBackToken();
    putCodeWithObject(IntermediateCode::RESET, (Virtual*) 0); 
  }
}


//--------------------------
  void Parser::parseSleep ()
//--------------------------
{
  readNextToken();
  parseExpression();
  putCodeWithID(IntermediateCode::CAST, NumericType::INT);
  putCode(IntermediateCode::SLEEP);  
}


//-------------------------
  void Parser::parseQuit ()
//-------------------------
{
  readNextToken();
  parseExpression();
  putCodeWithID(IntermediateCode::CAST, NumericType::INT);
  putCode(IntermediateCode::QUIT);
}


//----------------------------
  void Parser::parseCapture ()
//----------------------------
{
  Variable* capvar; 

  if ( readNextToken() == Token::SEMICOLON )
  {
    Function* function = expectInsideFunction(Token::CAPTURE);
    capvar = function->getCapVar();
    putBackToken();
  }
  else
  {
    expect(Token::IDENTIFIER);
    Identifier* identifier = parseIdentifier();
    expect(identifier, VAR_IDENT);
    capvar = dynamic_cast<Variable*>(identifier->getValue().get());
    checkType(capvar->getType(), &StandardType::BYTE );
  }

  capvar->use();
  putCodeWithObject(IntermediateCode::LOAD_PTR, capvar); 
  putCode(IntermediateCode::CAPTURE); 
}


//----------------------------
  void Parser::parseRelease ()
//----------------------------
{
  Variable* capvar; 

  if ( readNextToken() == Token::SEMICOLON )
  {
    Function* function = expectInsideFunction(Token::RELEASE);
    capvar = function->getCapVar();
    putBackToken();
  }
  else
  {
    expect(Token::IDENTIFIER);
    Identifier* identifier = parseIdentifier();
    expect(identifier, VAR_IDENT);
    capvar = dynamic_cast<Variable*>(identifier->getValue().get());
    checkType(capvar->getType(), &StandardType::BYTE );
  }

  capvar->use();
  putCodeWithObject(IntermediateCode::LOAD_PTR, capvar); 
  putCode(IntermediateCode::RELEASE); 
}


//-------------------------
  void Parser::parseWait ()
//-------------------------
{
  unsigned mark = seq.top()->mark();
  putCode(IntermediateCode::YIELD);  
  readNextToken();
  parseExpression();
  putCodeWithID(IntermediateCode::CAST, NumericType::INT);
  putCodeWithID(IntermediateCode::BRANCH_IF_ZERO, mark);  
}

    
//-------------------------
  void Parser::parseLoop ()
//-------------------------
{  
  breakMarks.push(seq.top()->getMark());

  unsigned mark = seq.top()->mark();
  continueMarks.push(mark);
  readNextToken();
  parseInstruction ();
  putCodeWithID(IntermediateCode::GOTO, mark);

  seq.top()->mark(breakMarks.top());
  breakMarks.pop();
  continueMarks.pop();
}


//--------------------------
  void Parser::parseWhile ()
//--------------------------
{
  // while something
  // {
  // ...
  // }

  unsigned instNo = program->getLoggedInstructionNo();

  breakMarks.push(seq.top()->getMark());
  continueMarks.push(seq.top()->getMark());

  unsigned beforeCondition = seq.top()->getMark();
  putCodeWithID(IntermediateCode::GOTO, beforeCondition);  

  IntermediateCodeSequence condition;
  enterCodeSeq(&condition);
  readNextToken();
  parseExpression();
  putCodeWithID(IntermediateCode::CAST, NumericType::INT);
  leaveCodeSeq();

  unsigned beforeLoop = seq.top()->mark();
  readNextToken();
  parseInstruction();
  putCodeWithID(IntermediateCode::ENTER_INSTRUCTION, instNo);
  seq.top()->mark(continueMarks.top());
  
  seq.top()->mark(beforeCondition);
  chainCodeSeq(&condition);
  putImmLoadValueCode(0);
  putCode(IntermediateCode::EQUAL);
  putCodeWithID(IntermediateCode::BRANCH_IF_ZERO, beforeLoop);  

  seq.top()->mark(breakMarks.top());
  breakMarks.pop();
  continueMarks.pop();
}


//-----------------------
  void Parser::parseDo ()
//-----------------------
{
  // do
  // {
  // ...
  // } while something;
  
  breakMarks.push(seq.top()->getMark());
  continueMarks.push(seq.top()->getMark());

  unsigned beforeLoop = seq.top()->mark();
  readNextToken();
  parseInstruction();
  seq.top()->mark(continueMarks.top());

  markExecutableLine();
  readNextToken(Token::WHILE);
  readNextToken();
  parseExpression();
  putCodeWithID(IntermediateCode::CAST, NumericType::INT);
  putImmLoadValueCode(0);
  putCode(IntermediateCode::EQUAL);
  putCodeWithID(IntermediateCode::BRANCH_IF_ZERO, beforeLoop);  

  seq.top()->mark(breakMarks.top());
  breakMarks.pop();
  continueMarks.pop();
}


//-----------------------
  void Parser::parseIf ()
//-----------------------
{
  // if something
  // {
  //  ...
  // }
  // else
  // {
  //  ...
  // }

  unsigned afterIf = seq.top()->getMark();

  readNextToken();
  parseExpression();
  putCodeWithID(IntermediateCode::CAST, NumericType::INT);
  putCodeWithID(IntermediateCode::BRANCH_IF_ZERO, afterIf);  

  readNextToken();
  parseInstruction();

  savePosition();

  if ( input->getToken().id != Token::RIGHT_BRACE )
    readNextToken(Token::SEMICOLON);

  if ( readNextToken() == Token::ELSE )
  {
    unsigned afterElse = seq.top()->getMark();
  	putCodeWithID(IntermediateCode::GOTO, afterElse);
    seq.top()->mark(afterIf);
    readNextToken();
    parseInstruction();
    seq.top()->mark(afterElse);     
  }
  else
  {
    restorePosition();
    seq.top()->mark(afterIf);  
  }   
}


//-----------------------
  void Parser::parseOn ()
//-----------------------
{
}


//------------------------
  void Parser::parseFor ()
//------------------------
{
  // for i = 0 ... <100 step something
  // {
  //   ....
  // }

  unsigned instNo = program->getLoggedInstructionNo();

  IntermediateCodeSequence loopVarAccessSeq;
  IntermediateCodeSequence initSeq;
  IntermediateCodeSequence compareSeq;
  IntermediateCodeSequence stepSeq;
  IntermediateCodeSequence loopActionSeq;

  bool fastStep;
  bool isIndirect = false;
  Identifier* identifier;
  Identified* identified;
  Variable*   variable; 

  breakMarks.push(seq.top()->getMark());
  continueMarks.push(seq.top()->getMark());

  // *** loopVarAccessSeq ***

  readNextToken();
  identifier = parseIdentifier();
  expect(identifier, VAR_IDENT);

  enterCodeSeq(&loopVarAccessSeq);
  identified = dynamic_cast<Identified*>(identifier->getValue().get());
  identified->use();
  variable = dynamic_cast<Variable*>(parseDataAccess(dynamic_cast<Data*>(identified), isIndirect, false));
  leaveCodeSeq();
  expectNumericType(variable->getType());


  // *** initSeq ***

  readNextToken(Token::ASSIGN);
  enterCodeSeq(&initSeq);
  readNextToken();
  parseExpression();
  putCodeWithID(IntermediateCode::CAST, (dynamic_cast<NumericType*>(variable->getType()))->getForm());
  leaveCodeSeq();


  // *** compareSeq ***

  readNextToken(Token::ELLIPSE);
  IntermediateCode::OPERATION opcode;
 
  switch ( readNextToken() )
  {
    case Token::HIGHER     : opcode = IntermediateCode::LOWERSAME; break;
    case Token::LOWER      : opcode = IntermediateCode::HIGHERSAME; break;
    case Token::HIGHERSAME : opcode = IntermediateCode::LOWER; break;
    case Token::LOWERSAME  : opcode = IntermediateCode::HIGHER; break;
    
    default :  
     
      opcode = IntermediateCode::HIGHER; 
      putBackToken();
      break;
  }
  
  enterCodeSeq(&compareSeq);
  readNextToken();
  parseExpression();
  putCodeWithID(IntermediateCode::CAST, (dynamic_cast<NumericType*>(variable->getType()))->getForm());
  leaveCodeSeq();


  // *** stepSeq ***

  enterCodeSeq(&stepSeq);

  if ( readNextToken() == Token::STEP )
  {
    fastStep = false;
    readNextToken();
    parseExpression();
  }
  else
  {
    fastStep = false;   
    putBackToken();
  }

  leaveCodeSeq();


  // *** loopActionSeq ***

  enterCodeSeq(&loopActionSeq);
  readNextToken();
  parseInstruction();
  seq.top()->mark(continueMarks.top());
  leaveCodeSeq();


  // *** build whole sequence ***

  if ( fastStep )
  {
    unsigned compareMark = seq.top()->getMark();

    chainCodeSeq(&initSeq);   
    chainCodeSeq(&loopVarAccessSeq);   
    putCodeWithObject(isIndirect? IntermediateCode::R_STORE : IntermediateCode::STORE, variable);
    chainCodeSeq(&loopVarAccessSeq);   
    putCodeWithObject(isIndirect? IntermediateCode::R_LOAD : IntermediateCode::LOAD, variable);

    putCodeWithID(IntermediateCode::GOTO, compareMark);

    unsigned loopStartMark = seq.top()->mark();
    chainCodeSeq(&loopActionSeq);   
    putCodeWithID(IntermediateCode::ENTER_INSTRUCTION, instNo);
    chainCodeSeq(&loopVarAccessSeq);   
    putCodeWithObject(isIndirect? IntermediateCode::R_INC_LOAD : IntermediateCode::INC_LOAD, variable);
    seq.top()->mark(compareMark);
    chainCodeSeq(&compareSeq);   
    putCode(opcode);
    putCodeWithID(IntermediateCode::BRANCH_IF_ZERO, loopStartMark);  
  }
  else
  {
    unsigned storeMark = seq.top()->getMark();

    chainCodeSeq(&initSeq);   
    putCodeWithID(IntermediateCode::GOTO, storeMark);

    unsigned loopStartMark = seq.top()->mark();
    chainCodeSeq(&loopActionSeq);   

    putCodeWithID(IntermediateCode::ENTER_INSTRUCTION, instNo);
    chainCodeSeq(&loopVarAccessSeq);   
    putCodeWithObject(isIndirect? IntermediateCode::R_LOAD : IntermediateCode::LOAD, variable);

    if ( stepSeq.isEmpty() )
      putImmLoadValueCode(1);
    else
      chainCodeSeq(&stepSeq);   
    putCode(IntermediateCode::ADD);
    putCodeWithID(IntermediateCode::CAST, (dynamic_cast<NumericType*>(variable->getType()))->getForm());
    seq.top()->mark(storeMark);
    chainCodeSeq(&loopVarAccessSeq);   
    putCodeWithObject(isIndirect? IntermediateCode::R_STORE : IntermediateCode::STORE, variable);

    chainCodeSeq(&loopVarAccessSeq);   
    putCodeWithObject(isIndirect? IntermediateCode::R_LOAD : IntermediateCode::LOAD, variable);
    chainCodeSeq(&compareSeq);   
    putCode(opcode);
    putCodeWithID(IntermediateCode::BRANCH_IF_ZERO, loopStartMark);  
  }
  
  
  seq.top()->mark(breakMarks.top());
  breakMarks.pop();
  continueMarks.pop();
}

