/********************************************************************

    project C-Control II Compiler

    project.cpp

********************************************************************/


#include "project.h"


//---------------------------------------------
  Project::Project ( const std::string& _name )
//---------------------------------------------
{
  name = _name;
  removeAll();
  errLog = 0;
}


//--------------------
  Project::~Project ()
//--------------------
{
  delete errLog;
}


//-------------------------------------------------
  void Project::rename ( const std::string& _name )
//-------------------------------------------------
{
  name = _name;
}


//---------------------------------------------------------
  bool Project::hasModule ( const std::string& modulename )
//---------------------------------------------------------
{
  for ( ModuleIterator iterator = moduleList.begin(); iterator != moduleList.end(); ++iterator )
    if ( (*iterator)->getName() == modulename )
      return true;
  
  return false;
}


//----------------------------------------------------------
  bool Project::isGoodName ( const std::string& modulename )
//----------------------------------------------------------
{
  std::istrstream input(modulename.c_str());
  ScanResult output;

  scanner.run(&input, &output);
  output.gotoToken(0);

  return ( output.getNumberOfTokens() == 2 ) && ( output.getToken().id == Token::IDENTIFIER );
}


//---------------------------------------------------------------------------------------------------------------
  Project::ModuleHandle Project::add ( const std::string& modulename, const std::string& filename, char* strbuf )
//---------------------------------------------------------------------------------------------------------------
{
  if ( hasModule(modulename) )
    return (ModuleHandle) ERR_DUPLICATED_MODULE_NAME;

  if ( !isGoodName(modulename) )
    return (ModuleHandle) ERR_BAD_MODULE_NAME;    

  ModuleHandle handle = nextFreeModuleHandle++;
  Module* module = strbuf ? new Module(strbuf, handle) : new Module(filename, handle);
  ModuleIterator iterator = moduleList.insert(moduleList.end(), module);
  moduleHandleMap.insert(std::pair<ModuleHandle, ModuleIterator>(handle, iterator));

  program.resetScope();
  program.declare(1, modulename, module, MODULE_IDENT);
  
  return handle;
}
    

//----------------------------------------------------------------
  bool Project::move ( ModuleHandle hModule, ModuleHandle hOther )
//----------------------------------------------------------------
{
  ModuleHandleMap::iterator map_iter_module = moduleHandleMap.find(hModule);
  ModuleHandleMap::iterator map_iter_other  = moduleHandleMap.find(hOther);

  if ( (map_iter_module == moduleHandleMap.end()) || (map_iter_other == moduleHandleMap.end()) )
    return false;

  ModuleIterator list_iter_module = (*map_iter_module).second;
  ModuleIterator list_iter_other  = (*map_iter_other).second;
  
  Module* module = *list_iter_module;
 
  moduleList.erase(list_iter_module);
  list_iter_module = moduleList.insert(list_iter_other, module);
  (*map_iter_module).second = list_iter_module;

  return true;  
}


//--------------------------
  void Project::removeAll ()
//--------------------------
{
  for ( ModuleIterator iterator = moduleList.begin(); iterator != moduleList.end(); ++iterator )
    (*iterator)->destroy();

  moduleHandleMap.clear();
  moduleList.clear();

  nextFreeModuleHandle = FIRST_FREE_MODULE_HANDLE;  
}


//---------------------------------------------
  bool Project::remove ( ModuleHandle hModule )
//---------------------------------------------
{
  ModuleHandleMap::iterator map_iter_module = moduleHandleMap.find(hModule);

  if ( map_iter_module == moduleHandleMap.end() )
    return false;

  ModuleIterator list_iter_module  = (*map_iter_module).second;
  (*list_iter_module)->destroy();
  moduleHandleMap.erase(map_iter_module);
  moduleList.erase(list_iter_module);

  return true;  
}



//-----------------------------------------
  long Project::compile ( Module* module  )
//-----------------------------------------
{
  long errCount;
   
  module->reset();
  module->openSource();
  //xxx on fail?
  scanner.run(module->getSource(), module->getTokens());
  module->closeSource();

  program.resetScope();
  program.enterScope(module->getIdentifier());

  errCount = parser.run(&program, module, errLog, module->getSourceName());
  module->clearTokens();

  if ( errCount ) 
    program.cancelInstructionLogging();

  return errCount;
}


//---------------------------------------------
  long Project::compile ( ModuleHandle hModule )
//----------------------------------------------
{
  ModuleHandleMap::iterator map_iterator = moduleHandleMap.find(hModule);

  if ( map_iterator == moduleHandleMap.end() )
    return false;

  ModuleIterator list_iterator = (*map_iterator).second;
  Module* module = *list_iterator;

  if ( !openErrLog () )
    return -1;

  long errCount = compile(module);

  (*errLog) << '\x0';

  return errCount;
}


//---------------------------------------------
  long Project::compile ( ModuleIterator last )
//---------------------------------------------
{
  long errCount = 0;

  if ( !openErrLog () )
    return -1;

  for ( ModuleIterator iterator = moduleList.begin(); iterator != last; ++iterator )
  {
    Module* module = *iterator;
    errCount += compile(module);
  }

  (*errLog) << '\x0';
  return errCount;
}


//----------------------
  long Project::build ()
//----------------------
{
  beginBuild();
  long numErrors = compile(moduleList.end());

  if ( 0 == numErrors )
    finishBuild();
  
  return numErrors;  
}


//---------------------------
  void Project::beginBuild ()
//---------------------------
{
  program.beginBuild();
}


//----------------------------
  bool Project::finishBuild ()
//----------------------------
{
  return openErrLog() ? program.finishBuild(errLog) : false;
}



//---------------------------------------------------
  Module* Project::getModule ( ModuleHandle hModule )
//---------------------------------------------------
{
  ModuleHandleMap::iterator map_iterator = moduleHandleMap.find(hModule);

  if ( map_iterator == moduleHandleMap.end() )
    return 0;

  return *(*map_iterator).second; 
}


//-------------------------------------------------------
  void Project::setErrLog ( const std::string& filename )
//-------------------------------------------------------
{
  errLogFileName = filename;
  errLogBuf = 0;
  errLogBufLength = 0;
}


//-------------------------------------------------
  void Project::setErrLog ( char* buf, int length )
//-------------------------------------------------
{
  errLogFileName = "";
  errLogBuf = buf;
  errLogBufLength = length;
}


//---------------------------
  bool Project::openErrLog ()
//---------------------------
{
  delete errLog;

  if ( errLogBuf )
    errLog = dynamic_cast<std::ostream*> (new std::ostrstream(errLogBuf, errLogBufLength));
  else
    errLog = dynamic_cast<std::ostream*> (new std::ofstream(errLogFileName.c_str()));

  return !errLog->fail();
}


//------------------------------------------------------------------------------------
  bool Project::writeOutputFile ( const std::string& filename, bool doWriteDebugFile )
//------------------------------------------------------------------------------------
{
  std::string outfilename = filename + ".vmc";

  std::ofstream os(outfilename.c_str());

  if ( doWriteDebugFile )
  {
    std::string dbgfilename = filename + ".dbg";
    std::ofstream dbgos(dbgfilename.c_str());
    return program.writeOutputFile(os, dbgos);
  }
  else
    return program.writeOutputFile(os);
}


//-----------------------------------------
  long Project::getRequiredVMCBufferSize ()
//-----------------------------------------
{
  return program.getRequiredVMCBufferSize();    
}


//-----------------------------------------------------
  void Project::fillVMCBuffer ( unsigned char* buffer )
//-----------------------------------------------------
{
  program.fillVMCBuffer(buffer);
}


//----------------------------------------------------------------------
  Identifier* Project::getIdentifierInContext ( const char* ident_c_str, 
                                                ModuleHandle hModule, 
                                                unsigned long lino )
//----------------------------------------------------------------------
{
  //--- get module context ---

  ModuleHandleMap::iterator map_iterator = moduleHandleMap.find(hModule);

  if ( map_iterator == moduleHandleMap.end() )
    return 0;

  ModuleIterator list_iterator = (*map_iterator).second;
  Module* module = *list_iterator;

  //--- scan and parse ident_c_str ---

  std::istrstream input(ident_c_str);
  ScanResult output;

  scanner.run(&input, &output);
  output.gotoToken(0);

  Identifier* identifier = module->getContext(lino);

  for ( int i = 0; i < output.getNumberOfTokens()-1; i++ )
  {
    if ( output.getToken().id != Token::IDENTIFIER )
      return 0;

    identifier = identifier->seekChild(output.getString());

    if ( !identifier )
      return 0;

    output.gotoNextToken();

    if ( output.getToken().id != Token::DOT )
      break;

    output.gotoNextToken();
  }

  return identifier;
}


//--------------------------------------------------------
  bool Project::getIdentDefPos ( const char* ident_c_str, 
                                 ModuleHandle* hModule_ptr, 
                                 unsigned long* lino_ptr )
//--------------------------------------------------------
{
  Identifier* identifier = getIdentifierInContext(ident_c_str, *hModule_ptr, *lino_ptr);

  if ( !identifier )
    return false;

  *lino_ptr = dynamic_cast<Identified*>(identifier->getValue().get())->getDeclarationLineNumber();
  
  while ( identifier->getParent() != program.getRootScope() )
    identifier = identifier->getParent();

  *hModule_ptr = dynamic_cast<Module*>(identifier->getValue().get())->getHandle();
    
  return true;
}


//------------------------------------------------------------------------------------------------------
  bool Project::getIdentCompletions ( const char* ident_c_str, ModuleHandle hModule, unsigned long lino, 
                                      char* completions, unsigned long length )
//------------------------------------------------------------------------------------------------------
{
  Identifier* identifier = getIdentifierInContext(ident_c_str, hModule, lino);

  if ( !identifier )
    return false;

  std::ostrstream output(completions, length);
  unsigned numCompletions = identifier->getNumChildren();
  Identifier* completionIdent = identifier->getFirstChild();

  for ( unsigned i = 0; i < numCompletions; i++ )
  {
    Identified* identified = dynamic_cast<Identified*>(completionIdent->getValue().get());

    if ( !identified->isSecret() )
    {
      output << completionIdent->getKey();

      switch ( identified->what() )
      {
         case MODULE_IDENT   : output << " (module)"; break;
         case TYPE_IDENT     : output << " (type)"; break;
         case FUNCTION_IDENT : output << " (function)"; break;
         case THREAD_IDENT   : output << " (thread)"; break;     
         case CONST_IDENT :
         case ARRAY_CONST_IDENT :
         case STRING_CONST_IDENT :
         case STRING_ARRAY_CONST_IDENT : output << " (constant)"; break;

         case VAR_IDENT : output << " (variable)"; break;     
 
         default : break;
      }
   
      output << std::endl;
    }

    completionIdent = identifier->getNextChild();
  }

  output << (char) '\x0';

  return true;
}


