/********************************************************************

    project C-Control II Compiler

    scanner.cpp

********************************************************************/


#include "scanner.h"
#include "text.h"

#include <memory.h>



//--- ScanResult ----------------------------------------------------


//-------------------------
  ScanResult::ScanResult ()
//-------------------------
{
  tokens = 0;
  floats = 0;
  strings = 0;

  free();
}


//--------------------------
  ScanResult::~ScanResult ()
//--------------------------
{
  free();
}

   
//-------------------------
  void ScanResult::init ()
//-------------------------
{
  free();

  tokenCapacity  = 16384;
  floatCapacity  = 32;
  stringCapacity = 1024;

  tokens = new Token[tokenCapacity];
  floats = new double[floatCapacity];
  strings = new std::string[stringCapacity];
}


//------------------------
  void ScanResult::free ()
//------------------------
{
  delete [] tokens;
  delete [] floats;
  delete [] strings;
  
  tokens = 0;
  floats = 0;
  strings = 0;

  tokenCapacity = 0;
  floatCapacity = 0;
  stringCapacity = 0;

  numberOfTokens = 0;
  numberOfFloats = 0;
  numberOfStrings = 0;

  iToken = 0;
  iFloat = 0;
  iString = 0;
}


//------------------------------------------
  void ScanResult::putToken ( Token::ID id )
//------------------------------------------
{
  if ( iToken == tokenCapacity )
  {
    unsigned long newCapacity = tokenCapacity * 2;
    Token* newPointer = new Token[newCapacity];

    while ( tokenCapacity-- )
      newPointer[tokenCapacity] = tokens[tokenCapacity];

    delete [] tokens;
    tokens = newPointer;
    tokenCapacity = newCapacity;
  }

  tokens[iToken++].id = id;
}


//------------------------------------------------------
  void ScanResult::putToken ( Token::ID id, long value )
//------------------------------------------------------
{
  putToken(id);
  tokens[iToken-1].value = value;
}


//-------------------------------------------------------------
  void ScanResult::putFloatToken ( Token::ID id, double value )
//-------------------------------------------------------------
{
  putToken(id);
  tokens[iToken-1].value = iFloat;

  if ( iFloat == floatCapacity )
  {
    unsigned long newCapacity = floatCapacity * 2;    
    double* newPointer = new double[newCapacity];

    while ( floatCapacity-- )
      newPointer[floatCapacity] = floats[floatCapacity];

    delete [] floats;
    floats = newPointer;
    floatCapacity = newCapacity;
  }
  floats[iFloat++] = value;
}


//---------------------------------------------------------------------------
  void ScanResult::putStringToken ( Token::ID id, std::string const& string )
//---------------------------------------------------------------------------
{
  putToken(id);
  tokens[iToken-1].value = iString;

  if ( iString == stringCapacity )
  {
    unsigned long newCapacity = stringCapacity * 2;
    std::string* newPointer = new std::string[newCapacity];

    while ( stringCapacity-- )
      newPointer[stringCapacity] = strings[stringCapacity];

    delete [] strings;
    strings = newPointer;
    stringCapacity = newCapacity;
  }
  strings[iString++] = string;
}


//--------------------------
  void ScanResult::finish ()
//--------------------------
{
  numberOfTokens  = iToken;
  numberOfStrings = iString;
  numberOfFloats  = iFloat;
}


//--- Scanner -------------------------------------------------------


//-----------------------------------------------------------
  unsigned short  Scanner::hash ( const std::string& string )
//-----------------------------------------------------------
{
  short result = 0;
  short i = string.length() - 1; 
 
  const short mask[16] = { 31, 109, 97, 17, 47, 71, 3, 19, 
                           23, 151, 139, 67, 73, 233, 211, 59 };

  do 
  {
    result ^= string[i];
    result <<= 1;
    result ^= mask[i & 0xF];
  } while ( --i >= 0 );


  return result % KEYWORD_HASH_TABLE_LENGTH;
}


//-------------------
  Scanner::Scanner ()
//-------------------
{
  int i;

  ch     = 0;
  input  = 0;
  output = 0;

  maxKeywordLength = 0;
  
  for ( i = 0; i < KEYWORD_HASH_TABLE_LENGTH; ++i )
    hashtable[i] = 0;

  for ( i = 0; i < sizeof(keywords)/sizeof(keywords[0]); ++i )
  {
    Token::ID      id        = (Token::ID)(Token::KEYWORDS_BEGIN + 1 + i);
    std::string    string    = Text::get(id);
    unsigned short hashValue = hash(string);

    if ( hashtable[hashValue] ) throw string;    // for DEBUG only 

    keywords[i].id = id; 
    keywords[i].string = string; 
    hashtable[hashValue] = i + 1;

    if ( string.length() > maxKeywordLength )
      maxKeywordLength = string.length();
  }
}


//---------------------------------------------------------------
  void Scanner::run ( std::istream* _input, ScanResult* _output )
//---------------------------------------------------------------
{
  ch     = 0;
  input  = _input;
  output = _output;

  output->init();

  skipWhites();

  while ( getChar() )
  {
    if ( isalpha(ch) || isChar('_') )
      scanKeywordOrIdentifier();
    else if ( isdigit(ch) )
      scanLiteralValue();
    else
      switch (ch)
      {
        case '"' : scanLiteralString(); break;
        case '\'': scanLiteralChar(); break;
        default: scanOperator ();
      }

    skipWhites();
  }

  output->putToken(Token::END_OF_FILE);
  output->finish();
}


//-------------------------------------------------------------
  Token::ID Scanner::getKeywordID ( const std::string& string )
//-------------------------------------------------------------
{
  if ( string.length() > maxKeywordLength )
    return Token::INVALID;

  unsigned short i = hashtable[hash(string)];

  if ( i )
  {
    if ( !string.compare(keywords[--i].string) )
      return keywords[i].id;
  }

  return Token::INVALID;
}


//----------------------------------------
  unsigned char Scanner::getControlChar ()
//----------------------------------------
{
  switch ( tolower(getChar()) )
  {
    case 'a' : ch = '\a'; break;
    case 'b' : ch = '\b'; break;
    case 'f' : ch = '\f'; break;
    case 'n' : ch = '\n'; break;
    case 'r' : ch = '\r'; break;
    case 't' : ch = '\t'; break;
    case '\'' : ch = '\''; break;
    case '\"' : ch = '\"'; break;
    case '\\' : ch = '\\'; break;
   

// 15.12.2000: skipWites() entfernt

    case 'x' : ch = (unsigned char) scanAndGetHexInteger(); /* skipWhites(); */ break; 

    default  : if ( isdigit(ch) ) 
               {
                 putBackChar();
                 ch = (unsigned char) scanAndGetDecInteger();
                 //skipWhites(); 
               }
               break;
  }

  return ch;
}


//---------------------------------------------------
  bool Scanner::isNextChar ( const unsigned char cc )
//---------------------------------------------------
{
  if ( cc == getChar() )
    return true;
  else
    putBackChar();

  return false;
}


//--------------------------
  bool Scanner::isNewLine ()
//--------------------------
{
  if ( 0xA == ch )
    return true;
  else if ( 0xD == ch )
  {  
    isNextChar(0xA);
    return true;
  }
  else
    return false;
}

 
//----------------------------------------------------
  bool Scanner::isNextCharI ( const unsigned char cc )
//----------------------------------------------------
{
  if ( toupper(cc) == toupper(getChar()) )
    return true;
  else
    putBackChar();

  return false;
}


//---------------------------
  void Scanner::skipWhites ()
//---------------------------
{
  while ( getChar() )
  {
    if ( isNewLine() )
      output->putToken(Token::LINE_FEED);

    if ( ch > ' ')
    { 
      putBackChar();
      break;
    }
  }
}


//-------------------------------
  void Scanner::invalidateChar ()
//-------------------------------
{
  char wrong[2] = {ch, 0};
  std::string string = wrong;
  output->putStringToken(Token::INVALID, string);  
}


//----------------------------------------
  void Scanner::scanKeywordOrIdentifier ()
//----------------------------------------
{
  std::string string = "";

  string += ch;
  
  while ( getChar() ) 
  {
    if ( isalpha(ch) || isdigit(ch) || isChar('_') )
      string += ch;
    else
    {
      putBackChar();
      break;
    }
  } 

  Token::ID id = getKeywordID(string);
  
  if ( Token::INVALID != id )
    output->putToken(id);
  else
    output->putStringToken(Token::IDENTIFIER, string);
}


//----------------------------
  void Scanner::scanComment ()
//----------------------------
{
  while ( getChar() )
  {
    if ( isNewLine() )
      output->putToken(Token::LINE_FEED);

    if ( isChar('*') && isNextChar('/') )
      break;
  }
}


//-----------------------------
  void Scanner::scanOperator ()
//-----------------------------
{
  switch ( ch )
  {
    case '.' : output->putToken(getDotOrEllipse()); break;
    case ';' : output->putToken(Token::SEMICOLON); break;

    case '=' : output->putToken( isNextChar('=') ? Token::EQUAL : Token::ASSIGN );
               break;

    case ',' : output->putToken(Token::COMMA); break;

    case '(' : output->putToken(Token::LEFT_PARENTHESIS); break;
    case ')' : output->putToken(Token::RIGHT_PARENTHESIS); break;

    case '+' : output->putToken(Token::PLUS); break;
    case '-' : output->putToken(Token::MINUS); break;
    case '*' : output->putToken(Token::MUL); break;

    case '/' : switch ( getChar() )
               {
                 case '/': scanCommentUntilEol (); break;
                 case '*': scanComment (); break;
                 default: putBackChar(); output->putToken(Token::DIV); break;
               }
               break;

    case '{' : output->putToken(Token::LEFT_BRACE); break;
    case '}' : output->putToken(Token::RIGHT_BRACE); break;
    case '[' : output->putToken(Token::LEFT_BRACKET); break;
    case ']' : output->putToken(Token::RIGHT_BRACKET); break;


    case '>' : output->putToken( isNextChar('=') ? Token::HIGHERSAME :
                                 isNextChar('>') ? Token::SHR : Token::HIGHER );
               break;

    case '<' : output->putToken( isNextChar('=') ? Token::LOWERSAME :
                                 isNextChar('<') ? Token::SHL : Token::LOWER );
               break;

    case '!' : switch ( getChar() )
               {
                 case '=': output->putToken(Token::NOTEQUAL); break;
                 case '&': output->putToken(Token::NAND); break;
                 case '|': output->putToken(Token::NOR); break;
                 default: putBackChar(); output->putToken(Token::NOT); break;
               }
               break;

    case '&' : output->putToken(Token::AND); break;
    case '|' : output->putToken(Token::OR); break;
    case '%' : output->putToken(Token::MOD); break;
    case '~' : output->putToken(Token::NOT); break;
    case '^' : output->putToken(Token::XOR); break;
    //case '?' : output->putToken(Token::QUESTIONMARK); break;
    //case ':' : output->putToken(Token::COLON); break;

    default  : invalidateChar(); break;
  }
}


//-------------------------------------
  Token::ID Scanner::getDotOrEllipse ()
//-------------------------------------
{
  if ( isNextChar('.') )
  {
    if ( isNextChar('.') )
      return Token::ELLIPSE;
    else
      putBackChar();
  } 

  return Token::DOT;
}


//---------------------------------
  void Scanner::scanLiteralValue ()
//---------------------------------
{
  if ( '0' == ch )
  {
    if ( isNextCharI('x') )
      output->putToken(Token::LITERAL_INT, (long)scanAndGetHexInteger());
    else if ( isNextCharI('b') )
      output->putToken(Token::LITERAL_INT, (long)scanAndGetBinInteger());
    else
      scanDecValue();
  }
  else
  {
    putBackChar();
    scanDecValue();
  }
}


//-----------------------------
  void Scanner::scanDecValue ()
//-----------------------------
{
  double value = scanAndGetDecInteger();

  if ( isNextChar('.') )
  {
    if ( Token::ELLIPSE == getDotOrEllipse() )
    {
      output->putToken(Token::LITERAL_INT, (long)value);
      output->putToken(Token::ELLIPSE);
    }
    else
    {
      double decimalFactor = 1.0;

      while ( getChar() )
      {
        if ( !isdigit(ch) )
        {
          output->putFloatToken(Token::LITERAL_FLOAT, value);
          putBackChar();
          break;
        }

        decimalFactor *= 0.1;
        value += decimalFactor * (ch - '0');
      }
    }
  }
  else
    output->putToken(Token::LITERAL_INT, (long)value);
}


//------------------------------------
  int Scanner::scanAndGetDecInteger ()
//------------------------------------
{
  int value = 0;

  while ( getChar() )
  {
    if ( !isdigit(ch) )
    {
      putBackChar();
      break;
    }

    value *= 10;
    value += ch - '0';
  }

  return value;
}


//------------------------------------
  int Scanner::scanAndGetHexInteger ()
//------------------------------------
{
  int  value = 0;
  int  digit;

  while ( getChar() )
  {
    if ( !isxdigit(ch) )
    {
      putBackChar();
      break;
    }

    if ( isdigit(ch) )
      digit = ch - '0';
    else
      digit = 10 + tolower(ch) - 'a';

    value <<= 4;
    value += digit;
  }

  return value;
}


//------------------------------------
  int Scanner::scanAndGetBinInteger ()
//------------------------------------
{
  int  value = 0;
  int  digit;

  while ( getChar() )
  {
    if ( (ch != '1') && (ch != '0') )
    {
      putBackChar();
      break;
    }

    digit = ch - '0';

    value <<= 1;
    value += digit;
  }

  return value;
}


//--------------------------------
  void Scanner::scanLiteralChar ()
//--------------------------------
{
  unsigned char c = getChar();

  if ( c )
  {
    if ( isChar('\\') )
      c = getControlChar();

    if ( isNextChar('\'') )
      output->putToken (Token::LITERAL_CHAR, (long)c);
  }
}


//----------------------------------
  void Scanner::scanLiteralString ()
//----------------------------------
{
  std::string string = "";

  while ( getChar() )
  {
    if ( isNewLine() )
    {
      output->putToken (Token::UNEXPECTED_END_OF_LINE);
      break;
    }

    if ( isChar('"') )
    {
      skipWhites();

      if ( isNextChar ('"')	)
        continue;
      else
        break;
    }
    string += isChar('\\') ? getControlChar() : ch;
  }

  if ( ch )
    output->putStringToken (Token::LITERAL_STRING, string);
}


//------------------------------------
  void Scanner::scanCommentUntilEol ()
//------------------------------------
{
  while ( getChar() )
  {
    if ( isNewLine() )
    {
      output->putToken(Token::LINE_FEED);
      break;
    }
  }
}





